# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Value network training tasks."""

import copy

import numpy as np

from trax import layers as tl
from trax.fastmath import numpy as jnp
from trax.supervised import training


class ValueTrainTask(training.TrainTask):
  """Task for value training."""

  def __init__(
      self,
      trajectory_batch_stream,
      optimizer,
      lr_schedule,
      advantage_estimator,
      model,
      target_model=None,
      target_scale=1.0,
      sync_at=(lambda step: step % 100 == 0),
      loss_layer=None,
      head_selector=(),
  ):
    """Initializes ValueTrainTask.

    Args:
      trajectory_batch_stream: Generator of trax.rl.task.TimeStepBatch.
      optimizer: Optimizer for network training.
      lr_schedule: Learning rate schedule for network training.
      advantage_estimator: Function
        (rewards, returns, values, dones) -> advantages, created by one of the
        functions from trax.rl.advantages.
      model: Model being trained, used to synchronize weights of the target
        model.
      target_model: Model for calculating TD targets. If `None`, use `model`.
      target_scale: Multiplier for the targets. Useful for rescaling the targets
        to a more reasonable range for model training.
      sync_at: Function step -> bool, indicating when to synchronize the target
        network with the trained network. This is necessary for training the
        network on bootstrapped targets, e.g. using TD-k.
      loss_layer: The value loss layer. The default is L2 loss.
      head_selector: Layer to apply to the network output to select the value
        head. Only needed in multitask training.
    """
    self._trajectory_batch_stream = trajectory_batch_stream
    self._advantage_estimator = advantage_estimator
    self._target_scale = target_scale

    self._synced = False
    def sync_also_on_initial_batch(step):
      return sync_at(step) or not self._synced
    self._sync_at = sync_also_on_initial_batch

    self._head_selector = head_selector

    def attach_head(model):
      return tl.Serial(model, self._head_selector)
    self._train_model = attach_head(model)
    if target_model is None:
      target_model = model
    # TODO(pkozakowski): Use target_model.clone() once it's implemented.
    self._target_model = attach_head(copy.deepcopy(target_model))

    # Count the steps, so we know when to synchronize the target network.
    self._step = 0
    def labeled_data():
      for trajectory_batch in self._trajectory_batch_stream:
        self._step += 1
        yield self.value_batch(trajectory_batch)
    sample_batch = self.value_batch(
        next(trajectory_batch_stream), shape_only=True
    )
    if loss_layer is None:
      loss_layer = tl.L2Loss()
    loss_layer = tl.Serial(head_selector, loss_layer)
    super().__init__(
        labeled_data(), loss_layer, optimizer,
        sample_batch=sample_batch,
        lr_schedule=lr_schedule,
        loss_name='value_loss',
    )

  @property
  def trajectory_batch_stream(self):
    return self._trajectory_batch_stream

  def _sync_target_model(self):
    self._target_model.weights = self._train_model.weights
    self._target_model.state = self._train_model.state
    self._synced = True

  def value_batch(self, trajectory_batch, shape_only=False):
    """Computes a value training batch based on a trajectory batch.

    Args:
      trajectory_batch: trax.rl.task.TimeStepBatch with a batch of trajectory
        slices. Elements should have shape (batch_size, seq_len, ...).
      shape_only: Whether to return dummy zero arrays of correct shape. Useful
        for initializing models.

    Returns:
      Triple (observations, targets, weights), where targets are the target
      values for network training and weights are used for masking in loss
      computation. Shapes:
      - observations: (batch_size, seq_len) + observation_shape
      - targets: (batch_size, seq_len)
      - weights: (batch_size, seq_len)
    """
    if self._sync_at(self._step) and not shape_only:
      self._sync_target_model()

    (batch_size, seq_len) = trajectory_batch.observation.shape[:2]
    assert trajectory_batch.action.shape[:2] == (batch_size, seq_len)
    assert trajectory_batch.mask.shape == (batch_size, seq_len)
    # Compute the value from the target network.
    values = np.array(self.value(trajectory_batch, shape_only=shape_only))
    assert values.shape == (batch_size, seq_len)
    # Compute the advantages - the TD errors of the target network.
    advantages = self._advantage_estimator(
        rewards=trajectory_batch.reward,
        returns=trajectory_batch.return_,
        dones=trajectory_batch.done,
        values=values,
        discount_mask=trajectory_batch.env_info.discount_mask,
    )
    adv_seq_len = advantages.shape[1]
    # The advantage sequence should be shorter by the margin. For more details,
    # see the comment in policy_tasks.PolicyTrainTask.policy_batch.
    assert adv_seq_len <= seq_len
    assert advantages.shape == (batch_size, adv_seq_len)
    # Compute the targets based on the target values and their TD errors. The
    # network gives perfect predictions when targets == values, so the
    # advantages are zero.
    targets = (values[:, :adv_seq_len] + advantages) * self._target_scale
    # Trim observations and the mask to match the target length.
    observations = trajectory_batch.observation[:, :adv_seq_len]
    mask = trajectory_batch.mask[:, :adv_seq_len]
    # Add a singleton depth dimension to the targets and the mask.
    targets = targets[:, :, None]
    mask = mask[:, :, None]
    return (observations, targets, mask)

  def value(self, trajectory_batch, shape_only=False):
    """Computes values of states in a given batch of trajectory slices.

    Can be passed as value_fn to PolicyTrainTask to implement a critic baseline
    for advantage calculation.

    Args:
      trajectory_batch: Batch of trajectory slices to compute values for.
      shape_only: Whether to return dummy zero arrays of correct shape. Useful
        for initializing models.

    Returns:
      Array of values of all states in `trajectory_batch`.
    """
    if shape_only:
      # The target model hasn't been initialized yet, and we are asked for the
      # initial, sample batch. Only shape matters here, so just return zeros.
      return np.zeros(trajectory_batch.observation.shape[:2])

    if not self._synced:
      self._sync_target_model()

    values = self._target_model(trajectory_batch.observation)
    # Squeeze the singleton depth axis.
    return np.squeeze(values, axis=-1) / self._target_scale


class ValueEvalTask(training.EvalTask):
  """Task for value evaluation."""

  def __init__(self, train_task, n_eval_batches=1, head_selector=()):
    """Initializes ValueEvalTask.

    Args:
      train_task: ValueTrainTask used to train the policy network.
      n_eval_batches: Number of batches per evaluation.
      head_selector: Layer to apply to the network output to select the value
        head. Only needed in multitask training.
    """
    labeled_data = map(
        train_task.value_batch, train_task.trajectory_batch_stream
    )
    metrics = [tl.L2Loss(), self.l1_loss]
    metric_names = ['value_l2', 'value_l1']
    # Select the appropriate head for evaluation.
    metrics = [tl.Serial(head_selector, metric) for metric in metrics]
    super().__init__(
        labeled_data, metrics,
        sample_batch=train_task.sample_batch,
        metric_names=metric_names,
        n_eval_batches=n_eval_batches,
    )

  @property
  def l1_loss(self):
    def loss(values, targets, weights):
      return jnp.sum(jnp.abs(values - targets) * weights) / jnp.sum(weights)
    return tl.Fn('L1Loss', loss)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A debugging decorator for TRAX input pipeline."""

import functools

from absl import logging
import gin


@gin.configurable(denylist=['f'])
def debug_pipeline(f, debug=False, method='pow', log_prefix=None):
  """Decorator for input pipeline generators that logs examples at intervals."""
  if not debug:
    return f

  assert method in ('pow', 'every')
  @functools.wraps(f)
  def wrapper(*args, **kwargs):
    count = 0
    prefix = log_prefix or f.__name__
    for example in f(*args, **kwargs):
      count += 1
      if method == 'every' or (method == 'pow' and (count & count - 1 == 0)):
        logging.info('%s example[%d] = %r', prefix, count, example)
      yield example

  return wrapper



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for trax.shapes."""
from absl.testing import absltest
import numpy as np

from trax import shapes
from trax.shapes import ShapeDtype


class ShapesTest(absltest.TestCase):

  def test_constructor_and_read_properties(self):
    sd = ShapeDtype((2, 3), np.int32)
    self.assertEqual(sd.shape, (2, 3))
    self.assertEqual(sd.dtype, np.int32)

  def test_default_dtype_is_float32(self):
    sd = ShapeDtype((2, 3))
    self.assertEqual(sd.shape, (2, 3))
    self.assertEqual(sd.dtype, np.float32)

  def test_signature_on_ndarray(self):
    array = np.array([[2, 3, 5, 7],
                      [11, 13, 17, 19]],
                     dtype=np.int16)
    sd = shapes.signature(array)
    self.assertEqual(sd.shape, (2, 4))
    self.assertEqual(sd.dtype, np.int16)

  def test_shape_dtype_repr(self):
    sd = ShapeDtype((2, 3))
    repr_string = '{}'.format(sd)
    self.assertEqual(repr_string,
                     "ShapeDtype{shape:(2, 3), dtype:<class 'numpy.float32'>}")

  def test_splice_signatures(self):
    sd1 = ShapeDtype((1,))
    sd2 = ShapeDtype((2,))
    sd3 = ShapeDtype((3,))
    sd4 = ShapeDtype((4,))
    sd5 = ShapeDtype((5,))

    # Signatures can be ShapeDtype instances, tuples of 2+ ShapeDtype instances,
    # or empty tuples.
    sig1 = sd1
    sig2 = (sd2, sd3, sd4)
    sig3 = ()
    sig4 = sd5
    spliced = shapes.splice_signatures(sig1, sig2, sig3, sig4)
    self.assertEqual(spliced, (sd1, sd2, sd3, sd4, sd5))

  def test_len_signature(self):
    """Signatures of all sizes should give correct length when asked."""
    x1 = np.array([1, 2, 3])
    x2 = np.array([10, 20, 30])
    inputs0 = ()
    inputs1 = x1  # NOT in a tuple
    inputs2 = (x1, x2)

    sig0 = shapes.signature(inputs0)
    sig1 = shapes.signature(inputs1)
    sig2 = shapes.signature(inputs2)

    # pylint: disable=g-generic-assert
    self.assertEqual(len(sig0), 0)
    self.assertEqual(len(sig1), 1)
    self.assertEqual(len(sig2), 2)
    # pylint: enable=g-generic-assert


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for Terraformer models."""

import functools

from absl.testing import absltest
from absl.testing import parameterized
import gin
import numpy as np

from trax import fastmath
from trax import layers as tl
from trax import shapes
from trax.layers import test_utils
from trax.models.research import terraformer


BACKENDS = [fastmath.Backend.JAX]


def short_name(b):
  if b == fastmath.Backend.JAX:
    return 'jax'
  else:
    return 'tf'


class TerraformerTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    gin.clear_config()

  def _lsh_self_attention_fn(self):
    return functools.partial(
        tl.LSHSelfAttention,
        attention_dropout=0.0,
        chunk_len=64,
        n_buckets=[32, 32],
        n_chunks_after=0,
        n_chunks_before=1,
        n_hashes=1,
        n_parallel_heads=1,
        predict_drop_len=128,
        predict_mem_len=1024,
    )

  def _timebin_self_attention_fn(self, use_reference_code=False):
    return functools.partial(
        tl.SelfAttention,
        attention_dropout=0.05,
        chunk_len=64,
        n_chunks_before=1,
        n_parallel_heads=1,
        use_reference_code=use_reference_code
    )

  @parameterized.named_parameters(
      [('_%s_efficient' % short_name(backend), backend, tl.SelfAttention, False)
       for backend in BACKENDS] +
      [('_%s_causal' % short_name(backend), backend, tl.CausalAttention, False)
       for backend in BACKENDS] +
      # NOTE: tl.SelfAttention is not currently working for this case.
      [('_%s_preembed' % short_name(backend), backend, tl.CausalAttention, True)
       for backend in BACKENDS])
  def test_terraformer_quick(self, backend, encoder_attention_type, preembed):
    with fastmath.use_backend(backend):
      vocab_size = 2
      input_vocab_size = None if preembed else vocab_size
      output_vocab_size = vocab_size if preembed else None
      max_len = 2

      model = terraformer.ConfigurableTerraformer(
          input_vocab_size,
          d_model=4,
          d_ff=4,
          n_encoder_layers=1,
          n_decoder_layers=1,
          n_heads=2,
          dropout=0.05,
          max_len=max_len,
          pos_type=None,
          ff_activation=tl.Relu,
          ff_use_sru=0,
          ff_chunk_size=2,
          mode='train',
          output_vocab_size=output_vocab_size,
          encoder_attention_type=encoder_attention_type,
      )

      if preembed:
        model_inputs = [np.ones((1, max_len, 3)).astype(np.float32),
                        np.ones((1, max_len)).astype(np.bool)]
      else:
        model_inputs = [np.ones((1, max_len)).astype(np.int32)]
      x = model_inputs + [np.ones((1, max_len)).astype(np.int32)]
      model.init(shapes.signature(x))

      logits, dec_toks = model(x)
      del dec_toks

      self.assertEqual(logits.shape, (1, max_len, vocab_size))

  def test_terraformer_deterministic_eval(self):
    with fastmath.use_backend(fastmath.Backend.JAX):
      vocab_size = 16
      d_model = 4
      batch_size = 2
      length = 5

      model_fn = functools.partial(
          terraformer.ConfigurableTerraformer,
          vocab_size,
          d_model=d_model,
          d_ff=16,
          n_encoder_layers=0,
          n_decoder_layers=1,
          n_heads=2,
          dropout=0.0,
          max_len=length*2,
          pos_type=None,
          encoder_attention_type=tl.Attention,
          encoder_decoder_attention_type=tl.CausalAttention,
      )

      inp = np.random.randint(vocab_size, size=(batch_size, length))
      out = np.zeros((batch_size, length), dtype=np.int32)

      test_utils.test_eval_is_deterministic((inp, out), model_fn)

  def test_terraformer_predict_equals_eval(self):
    with fastmath.use_backend(fastmath.Backend.JAX):
      vocab_size = 16
      d_model = 8
      batch_size = 1
      length = 5

      model_fn = functools.partial(
          terraformer.ConfigurableTerraformer,
          vocab_size,
          d_model=d_model,
          d_ff=16,
          n_encoder_layers=1,
          n_decoder_layers=1,
          n_heads=2,
          ff_use_sru=(1, 8),  # ? is SRU working?
          dropout=0.0,
          max_len=(length+7)*2,
          pos_type=None,
          reversible_encoder=True,
          n_decoder_attention_layers=1,
          encoder_attention_type=tl.Attention,
          encoder_decoder_attention_type=tl.CausalAttention,
      )

      # Token id of 0 indicates padding; and predict mode doesn't support it.
      inp = np.random.randint(1, vocab_size, size=(batch_size, length))
      inp[:, -2:] = 0
      out = np.zeros((batch_size, length), dtype=np.int32)

      test_utils.test_eval_equals_predict(
          (inp, out), model_fn, seq_axis=1, seq_tensor=-1, init_tokens=1)

  def test_terraformer_doubling(self):
    vocab_size = 2
    max_len = 2

    model = terraformer.ConfigurableTerraformer(
        vocab_size,
        d_model=8,
        d_ff=16,
        n_encoder_layers=1,
        n_decoder_layers=6,
        n_heads=2,
        dropout=0.05,
        max_len=max_len,
        pos_type=None,
        half_before_layer=2,
        double_after_layer=2,
        encoder_attention_type=tl.Attention,
        encoder_decoder_attention_type=tl.CausalAttention,
        mode='train',
    )

    x = [np.ones((1, max_len)).astype(np.int32),
         np.ones((1, max_len)).astype(np.int32)]
    model.init(shapes.signature(x))

    logits, dec_toks = model(x)
    del dec_toks

    self.assertEqual(logits.shape, (1, max_len, vocab_size))

  def test_terraformer_one_step(self):
    vocab_size = 32
    max_len = 256
    pos_axial = 16
    assert pos_axial * pos_axial == max_len

    chunk_len = 32

    # Since 2 * chunk_len * n_buckets should be max_len.
    n_buckets = max_len // (2 * chunk_len)

    lsh_self_attention = functools.partial(self._lsh_self_attention_fn(),
                                           chunk_len=chunk_len,
                                           n_buckets=n_buckets)

    timebin_self_attention = self._timebin_self_attention_fn()

    model = terraformer.ConfigurableTerraformer(
        vocab_size,
        d_model=32,
        d_ff=64,
        d_attention_key=64,
        d_attention_value=64,
        n_encoder_layers=2,
        n_decoder_layers=2,
        n_heads=2,
        dropout=0.05,
        max_len=max_len,
        encoder_attention_type=lsh_self_attention,
        encoder_decoder_attention_type=[timebin_self_attention,
                                        lsh_self_attention],
        pos_axial_shape=(pos_axial, pos_axial),
        pos_d_axial_embs=(64, 192),
        ff_activation=tl.Relu,
        ff_use_sru=0,
        ff_chunk_size=64,
        ff_sparsity=8,
        mode='train',
    )

    x = [np.ones((1, max_len)).astype(np.int32),
         np.ones((1, max_len)).astype(np.int32)]
    weights, state = model.init(shapes.signature(x))

    @fastmath.jit
    def mock_training_step(x, weights, state, rng):
      def compute_mock_loss(weights):
        logits_and_dec_toks, new_state = model.pure_fn(x, weights, state, rng)
        # This returns [logits, decoder tokens]
        logits = logits_and_dec_toks[0]
        loss = fastmath.numpy.mean(logits[..., 0])
        return loss, (new_state, logits)
      gradients, (new_state, logits) = fastmath.grad(
          compute_mock_loss, has_aux=True)(weights)
      new_weights = fastmath.nested_map_multiarg(
          lambda w, g: w - 1e-4 * g, weights, gradients)
      return new_weights, new_state, logits

    weights, state, logits = mock_training_step(
        x, weights, state, fastmath.random.get_prng(0))

    self.assertEqual(logits.shape, (1, max_len, vocab_size))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for trax.rl.space_serializer."""
import gin
import gym
import numpy as np
from tensorflow import test
from trax.rl import space_serializer


class BoxSpaceSerializerTest(test.TestCase):

  def _make_space_and_serializer(
      self, low=-10, high=10, shape=(2,),
      # Weird vocab_size to test that it doesn't only work with powers of 2.
      vocab_size=257,
      # Enough precision to represent float32s accurately.
      precision=4,
  ):
    gin.bind_parameter('BoxSpaceSerializer.precision', precision)
    space = gym.spaces.Box(low=low, high=high, shape=shape)
    serializer = space_serializer.create(space, vocab_size=vocab_size)
    return (space, serializer)

  def _sample_batch(self, space):
    return np.reshape(space.sample(), (1,) + space.shape)

  def test_representation_length(self):
    (space, serializer) = self._make_space_and_serializer()
    input_array = self._sample_batch(space)
    representation = serializer.serialize(input_array)
    self.assertEqual(
        representation.shape, (1, serializer.representation_length))

  def test_commutes(self):
    (space, serializer) = self._make_space_and_serializer()
    input_array = self._sample_batch(space)
    representation = serializer.serialize(input_array)
    output_array = serializer.deserialize(representation)
    # Testing till 5 decimals to reduce flakyness.
    np.testing.assert_array_almost_equal(input_array, output_array, decimal=5)

  def test_representation_changes(self):
    (space, serializer) = self._make_space_and_serializer()
    array1 = self._sample_batch(space)
    array2 = -array1
    (repr1, repr2) = tuple(map(serializer.serialize, (array1, array2)))
    self.assertFalse(np.array_equal(repr1, repr2))

  def test_bounds_space(self):
    gin.bind_parameter('BoxSpaceSerializer.max_range', (-10.0, 10.0))
    (_, serializer) = self._make_space_and_serializer(
        # Too wide range to represent, need to clip.
        low=-1e18, high=1e18,
        shape=(1,))
    input_array = np.array([[1.2345]])
    representation = serializer.serialize(input_array)
    output_array = serializer.deserialize(representation)
    np.testing.assert_array_almost_equal(input_array, output_array)

  def test_significance_map(self):
    (_, serializer) = self._make_space_and_serializer(shape=(2,))
    np.testing.assert_array_equal(
        serializer.significance_map, [0, 1, 2, 3, 0, 1, 2, 3])

  def test_serializes_boundaries(self):
    vocab_size = 256
    precision = 4
    (_, serializer) = self._make_space_and_serializer(
        low=-1, high=1, shape=(1,), vocab_size=vocab_size, precision=precision,
    )
    input_array = np.array([[-1, 1]])
    representation = serializer.serialize(input_array)
    np.testing.assert_array_equal(
        representation, [[0] * precision + [vocab_size - 1] * precision]
    )


class DiscreteSpaceSerializerTest(test.TestCase):

  def setUp(self):
    super().setUp()
    self._space = gym.spaces.Discrete(n=2)
    self._serializer = space_serializer.create(self._space, vocab_size=2)

  def _sample_batch(self):
    return np.reshape(self._space.sample(), (1,) + self._space.shape)

  def test_representation_length(self):
    input_array = self._sample_batch()
    representation = self._serializer.serialize(input_array)
    self.assertEqual(
        representation.shape, (1, self._serializer.representation_length))

  def test_commutes(self):
    input_array = self._sample_batch()
    representation = self._serializer.serialize(input_array)
    output_array = self._serializer.deserialize(representation)
    np.testing.assert_array_almost_equal(input_array, output_array)

  def test_representation_changes(self):
    array1 = self._sample_batch()
    array2 = 1 - array1
    (repr1, repr2) = tuple(map(self._serializer.serialize, (array1, array2)))
    self.assertFalse(np.array_equal(repr1, repr2))

  def test_significance_map(self):
    np.testing.assert_array_equal(self._serializer.significance_map, [0])


class MultiDiscreteSpaceSerializerTest(test.TestCase):

  def setUp(self):
    super().setUp()
    self._space = gym.spaces.MultiDiscrete(nvec=[2, 2])
    self._serializer = space_serializer.create(self._space, vocab_size=2)

  def _sample_batch(self):
    return np.reshape(self._space.sample(), (1,) + self._space.shape)

  def test_representation_length(self):
    input_array = self._sample_batch()
    representation = self._serializer.serialize(input_array)
    self.assertEqual(
        representation.shape, (1, self._serializer.representation_length))

  def test_commutes(self):
    input_array = self._sample_batch()
    representation = self._serializer.serialize(input_array)
    output_array = self._serializer.deserialize(representation)
    np.testing.assert_array_almost_equal(input_array, output_array)

  def test_representation_changes(self):
    array1 = self._sample_batch()
    array2 = 1 - array1
    (repr1, repr2) = tuple(map(self._serializer.serialize, (array1, array2)))
    self.assertFalse(np.array_equal(repr1, repr2))

  def test_significance_map(self):
    np.testing.assert_array_equal(self._serializer.significance_map, [0, 0])


if __name__ == '__main__':
  test.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for RL training."""

import functools


from absl.testing import absltest
from absl.testing import parameterized

from trax import layers as tl
from trax import models
from trax import optimizers as opt
from trax import test_utils
from trax.rl import actor_critic
from trax.rl import advantages
from trax.rl import task as rl_task
from trax.supervised import lr_schedules


class ActorCriticTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    test_utils.ensure_flag('test_tmpdir')

  def test_a2ctrainer_save_restore(self):
    """Check save and restore of A2C trainer."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=0,
                          max_steps=20)
    body = lambda mode: tl.Serial(tl.Dense(64), tl.Relu())
    policy_model = functools.partial(models.Policy, body=body)
    value_model = functools.partial(models.Value, body=body)
    tmp_dir = self.create_tempdir().full_path
    trainer1 = actor_critic.A2C(
        task,
        value_model=value_model,
        value_optimizer=opt.Adam,
        value_batch_size=2,
        value_train_steps_per_epoch=1,
        policy_model=policy_model,
        policy_optimizer=opt.Adam,
        policy_batch_size=2,
        policy_train_steps_per_epoch=2,
        n_trajectories_per_epoch=2,
        n_shared_layers=1,
        output_dir=tmp_dir)
    trainer1.run(2)
    self.assertEqual(trainer1.current_epoch, 2)
    self.assertEqual(trainer1._value_trainer.step, 2)
    self.assertEqual(trainer1._policy_trainer.step, 4)
    # Trainer 2 starts where trainer 1 stopped.
    trainer2 = actor_critic.A2C(
        task,
        value_model=value_model,
        value_optimizer=opt.Adam,
        value_batch_size=2,
        value_train_steps_per_epoch=1,
        policy_model=policy_model,
        policy_optimizer=opt.Adam,
        policy_batch_size=2,
        policy_train_steps_per_epoch=2,
        n_trajectories_per_epoch=2,
        n_shared_layers=1,
        output_dir=tmp_dir)
    trainer2.run(1)
    self.assertEqual(trainer2.current_epoch, 3)
    self.assertEqual(trainer2._value_trainer.step, 3)
    self.assertEqual(trainer2._policy_trainer.step, 6)
    trainer1.close()
    trainer2.close()

  def test_sanity_a2ctrainer_cartpole(self):
    """Test-runs a2c on cartpole."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=0,
                          max_steps=2)
    body = lambda mode: tl.Serial(tl.Dense(64), tl.Relu())
    policy_model = functools.partial(models.Policy, body=body)
    value_model = functools.partial(models.Value, body=body)
    lr = lambda: lr_schedules.multifactor(  # pylint: disable=g-long-lambda
        constant=1e-4, warmup_steps=100, factors='constant * linear_warmup')
    trainer = actor_critic.A2C(
        task,
        n_shared_layers=1,
        value_model=value_model,
        value_optimizer=opt.Adam,
        value_lr_schedule=lr,
        value_batch_size=2,
        value_train_steps_per_epoch=2,
        policy_model=policy_model,
        policy_optimizer=opt.Adam,
        policy_lr_schedule=lr,
        policy_batch_size=2,
        policy_train_steps_per_epoch=2,
        n_trajectories_per_epoch=2)
    trainer.run(2)
    self.assertEqual(2, trainer.current_epoch)

  def test_sanity_ppo_cartpole(self):
    """Run PPO and check whether it correctly runs for 2 epochs.s."""
    task = rl_task.RLTask(
        'CartPole-v1', initial_trajectories=0, max_steps=200)

    lr = lambda: lr_schedules.multifactor(  # pylint: disable=g-long-lambda
        constant=1e-3,
        warmup_steps=100,
        factors='constant * linear_warmup')

    body = lambda mode: tl.Serial(tl.Dense(64), tl.Relu())
    policy_model = functools.partial(models.Policy, body=body)
    value_model = functools.partial(models.Value, body=body)
    trainer = actor_critic.PPO(
        task,
        n_shared_layers=1,
        value_model=value_model,
        value_optimizer=opt.Adam,
        value_lr_schedule=lr,
        value_batch_size=128,
        value_train_steps_per_epoch=10,
        policy_model=policy_model,
        policy_optimizer=opt.Adam,
        policy_lr_schedule=lr,
        policy_batch_size=128,
        policy_train_steps_per_epoch=10,
        n_trajectories_per_epoch=10)

    trainer.run(2)
    self.assertEqual(2, trainer.current_epoch)

  def test_sanity_loopawr(self):
    """Test-runs LoopAWR."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=0, max_steps=2)
    body = lambda mode: tl.Serial(tl.Dense(64), tl.Relu())
    model_fn = functools.partial(models.PolicyAndValue, body=body)
    trainer = actor_critic.LoopAWR(
        task,
        model_fn,
        batch_size=2,
        network_eval_at=(lambda _: True),
        policy_n_steps_per_epoch=2,
        value_n_steps_per_epoch=2,
        n_trajectories_per_epoch=1,
        n_eval_episodes=1,
    )
    trainer.run(2)
    self.assertEqual(2, trainer.current_epoch)

  @parameterized.named_parameters(('default', None),
                                  ('thresholds', ((70, 1.0, 0), (90, 4.0, 0))))
  def test_sanity_awrtrainer_transformer_cartpole(self, thresholds):
    """Test-runs AWR on cartpole with Transformer."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=2,
                          max_steps=2)
    body = lambda mode: models.TransformerDecoder(  # pylint: disable=g-long-lambda
        d_model=2, d_ff=2, n_layers=1, n_heads=1, mode=mode)
    policy_model = functools.partial(models.Policy, body=body)
    value_model = functools.partial(models.Value, body=body)
    lr = lambda: lr_schedules.multifactor(  # pylint: disable=g-long-lambda
        constant=1e-2, warmup_steps=100, factors='constant * linear_warmup')
    trainer = actor_critic.AWR(
        task,
        thresholds=thresholds,
        n_shared_layers=0,
        max_slice_length=2,
        added_policy_slice_length=1,
        value_model=value_model,
        value_optimizer=opt.Adam,
        value_lr_schedule=lr,
        value_batch_size=2,
        value_train_steps_per_epoch=2,
        policy_model=policy_model,
        policy_optimizer=opt.Adam,
        policy_lr_schedule=lr,
        policy_batch_size=2,
        policy_train_steps_per_epoch=2,
        n_trajectories_per_epoch=1,
        n_eval_episodes=1)
    trainer.run(2)
    self.assertEqual(2, trainer.current_epoch)

  def test_sampling_awrtrainer_cartpole(self):
    """Test-runs AWR on cartpole with Transformer."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=0,
                          max_steps=20)
    body = lambda mode: tl.Serial(tl.Dense(2), tl.Relu())
    policy_model = functools.partial(models.Policy, body=body)
    value_model = functools.partial(models.Value, body=body)
    lr = lambda: lr_schedules.multifactor(  # pylint: disable=g-long-lambda
        constant=1e-2, warmup_steps=100, factors='constant * linear_warmup')
    trainer = actor_critic.SamplingAWR(
        task,
        n_shared_layers=0,
        added_policy_slice_length=1,
        value_model=value_model,
        value_optimizer=opt.Adam,
        value_lr_schedule=lr,
        value_batch_size=2,
        value_train_steps_per_epoch=2,
        policy_model=policy_model,
        policy_optimizer=opt.Adam,
        policy_lr_schedule=lr,
        policy_batch_size=2,
        policy_train_steps_per_epoch=2,
        n_trajectories_per_epoch=2,
        advantage_estimator=advantages.monte_carlo,
        advantage_normalization=False,
        q_value_n_samples=3,
        q_value_aggregate='max',
        reweight=False,
    )
    trainer.run(1)
    self.assertEqual(1, trainer.current_epoch)

  def test_sampling_awrtrainer_cartpole_sample_all_discrete(self):
    """Test-runs AWR on cartpole with Transformer, n_actions = n_samples."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=0,
                          max_steps=20)
    body = lambda mode: tl.Serial(tl.Dense(2), tl.Relu())
    policy_model = functools.partial(models.Policy, body=body)
    value_model = functools.partial(models.Value, body=body)
    lr = lambda: lr_schedules.multifactor(  # pylint: disable=g-long-lambda
        constant=1e-2, warmup_steps=100, factors='constant * linear_warmup')
    trainer = actor_critic.SamplingAWR(
        task,
        n_shared_layers=0,
        added_policy_slice_length=1,
        value_model=value_model,
        value_optimizer=opt.Adam,
        value_lr_schedule=lr,
        value_batch_size=2,
        value_train_steps_per_epoch=2,
        policy_model=policy_model,
        policy_optimizer=opt.Adam,
        policy_lr_schedule=lr,
        policy_batch_size=2,
        policy_train_steps_per_epoch=2,
        n_trajectories_per_epoch=2,
        advantage_estimator=advantages.monte_carlo,
        advantage_normalization=False,
        q_value_n_samples=2,
        q_value_aggregate='max',
        reweight=False,
    )
    trainer.run(1)
    self.assertEqual(1, trainer.current_epoch)

  def test_sampling_awrtrainer_mountain_acr(self):
    """Test-runs Sampling AWR on MountainCarContinuous."""
    task = rl_task.RLTask('MountainCarContinuous-v0', initial_trajectories=0,
                          max_steps=2)
    body = lambda mode: tl.Serial(tl.Dense(2), tl.Relu())
    policy_model = functools.partial(models.Policy, body=body)
    value_model = functools.partial(models.Value, body=body)
    lr = lambda: lr_schedules.multifactor(  # pylint: disable=g-long-lambda
        constant=1e-2, warmup_steps=100, factors='constant * linear_warmup')
    trainer = actor_critic.SamplingAWR(
        task,
        n_shared_layers=0,
        added_policy_slice_length=1,
        value_model=value_model,
        value_optimizer=opt.Adam,
        value_lr_schedule=lr,
        value_batch_size=2,
        value_train_steps_per_epoch=2,
        policy_model=policy_model,
        policy_optimizer=opt.Adam,
        policy_lr_schedule=lr,
        policy_batch_size=2,
        policy_train_steps_per_epoch=2,
        n_trajectories_per_epoch=2,
        advantage_estimator=advantages.monte_carlo,
        advantage_normalization=False,
        q_value_n_samples=3,
    )
    trainer.run(1)
    self.assertEqual(1, trainer.current_epoch)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for trax.layers.research.position_encodings."""

import functools
import absl.testing.absltest as unittest
import numpy as np
import parameterized

from trax import fastmath
import trax.layers.research.position_encodings as pe


@parameterized.parameterized_class([
    # {'Encoding': pe.FixedBasePositionalEncoding},
    {'Encoding': pe.InfinitePositionalEncoding},
    {'Encoding': functools.partial(
        pe.InfinitePositionalEncoding, affine=False)},
    {'Encoding': functools.partial(
        pe.TimeBinPositionalEncoding, time_bin_length=5)},
])
class PositionEncodingsTest(unittest.TestCase):
  """Position encodings conform to the position encodings protocol."""

  @parameterized.parameterized.expand([
      (1, 100, 8),  # typical
      (1, 1, 8),  # short
      (1, 100, 1),  # narrow
      (2, 100, 8),  # batched
  ])
  def test_training(self, n, t, c):
    encoding = self.Encoding()
    input_ntc = np.random.randn(n, t, c)
    encoding.init(input_ntc)
    output_ntc = encoding(input_ntc)
    self.assertEqual(output_ntc.shape, input_ntc.shape)
    self.assertTrue(np.not_equal(output_ntc, input_ntc).any())

  @parameterized.parameterized.expand([
      (1, 100, 8),  # typical
      (1, 100, 1),  # narrow
      (2, 100, 8),  # batched
  ])
  def test_inference(self, n, t, c):
    # Get the eval mode outputs:
    encoding = self.Encoding(mode='eval')
    input_ntc = np.random.randn(n, t, c)
    rng = fastmath.random.get_prng(1234)
    encoding.init(input_ntc, rng=rng)
    output_ntc = encoding(input_ntc)

    is_random = self.Encoding == pe.InfinitePositionalEncoding

    # Get the predict mode outputs:
    encoding_pred = self.Encoding(mode='predict')
    encoding_pred.init(input_ntc[:, 0:1, :], rng=rng)
    output_ntc0 = encoding_pred(input_ntc[:, 0:1, :])
    if not is_random:
      np.testing.assert_allclose(output_ntc0, output_ntc[:, 0:1, :], atol=1e-4)

    output_ntc1 = encoding_pred(input_ntc[:, 1:2, :])
    if not is_random:
      np.testing.assert_allclose(output_ntc1, output_ntc[:, 1:2, :], atol=1e-4)

    output_ntc2 = encoding_pred(input_ntc[:, 2:3, :])
    if not is_random:
      np.testing.assert_allclose(output_ntc2, output_ntc[:, 2:3, :], atol=1e-4)


class SinCosEncodingsTest(unittest.TestCase):
  """Position encodings conform to the position encodings protocol."""

  @parameterized.parameterized.expand([
      (1, 100, 8),  # typical
      (1, 1, 8),  # short
      (2, 100, 8),  # batched
  ])
  def test_training(self, n, t, c):
    encoding = pe.SinCosPositionalEncoding()
    input_ntc = np.random.randn(n, t, c)
    encoding.init(input_ntc)
    output_ntc = encoding(input_ntc)
    self.assertEqual(output_ntc.shape, input_ntc.shape)


if __name__ == '__main__':
  unittest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Implementation of the improved Neural GPU (NGPU)."""

from trax import layers as tl
from trax.fastmath import numpy as jnp


# TODO(ddohan): Combinator to add saturation costs to loss
def SaturationCost(x, limit=0.9):
  return jnp.minimum(0, jnp.abs(x) - limit)


def DiagonalGate():
  """Split channels in 3 parts. Shifts 1st and 3rd sections to left/right."""

  def f(x):  # pylint: disable=invalid-name
    # x : [batch, 1, length, depth]
    x = jnp.pad(x, [(0, 0), (0, 0), (1, 1), (0, 0)],
                mode='constant', constant_values=0.0)
    depth = x.shape[-1] // 3
    assert 3 * depth == x.shape[-1], ('Depth must be divisible by 3', depth,
                                      x.shape)
    xs = [
        x[:, :, :-2, :depth], x[:, :, 1:-1, depth:2 * depth],
        x[:, :, 2:, 2 * depth:3 * depth]
    ]
    return jnp.concatenate(xs, axis=3)
  return tl.Fn('DiagonalGate', f)


def ConvDiagonalGRU(units, kernel_size=(3, 3)):
  """Build convolutional GRU with diagonal gating as in ImprovedNGPU."""

  def BuildConv():
    return tl.Conv(filters=units, kernel_size=kernel_size, padding='SAME')

  return tl.GeneralGRUCell(
      candidate_transform=BuildConv,
      memory_transform_fn=DiagonalGate,
      gate_nonlinearity=tl.HardSigmoid,
      candidate_nonlinearity=tl.HardTanh)


def NeuralGPU(d_feature=96, steps=16, vocab_size=2, mode='train'):
  """Implementation of Neural GPU: https://arxiv.org/abs/1702.08727.

  Args:
    d_feature: Number of memory channels (dimensionality of feature embedding).
    steps: Number of times depthwise recurrence steps.
    vocab_size: Vocabulary size.
    mode: Whether we are training or evaluating or doing inference.

  Returns:
    A NeuralGPU Stax model.
  """
  del mode

  core = ConvDiagonalGRU(units=d_feature)
  return tl.Serial(
      tl.Embedding(vocab_size=vocab_size, d_feature=d_feature),
      [core] * steps,
      tl.Dense(vocab_size),
  )



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Write Summaries from JAX for use with Tensorboard.

See jaxboard_demo.py for example usage.
"""
import io
import struct
import time
import warnings
import wave
import matplotlib as mpl
# Necessary to prevent attempted Tk import:
with warnings.catch_warnings():
  warnings.simplefilter('ignore')
  mpl.use('Agg')
# pylint: disable=g-import-not-at-top
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf

# pylint: disable=g-direct-tensorflow-import
from tensorflow.core.util import event_pb2
from tensorflow.python.summary.writer.event_file_writer import EventFileWriter
# pylint: enable=g-direct-tensorflow-import


def _pack_images(images, rows, cols):
  """Helper utility to make a tiled field of images from numpy arrays.

  Args:
    images: Image tensor in shape [N, W, H, C].
    rows: Number of images per row in tiled image.
    cols: Number of images per column in tiled image.

  Returns:
    A tiled image of shape [W * rows, H * cols, C].
    Truncates incomplete rows.
  """
  shape = np.shape(images)
  width, height, depth = shape[-3:]
  images = np.reshape(images, (-1, width, height, depth))
  batch = np.shape(images)[0]
  rows = np.minimum(rows, batch)
  cols = np.minimum(batch // rows, cols)
  images = images[:rows * cols]
  images = np.reshape(images, (rows, cols, width, height, depth))
  images = np.transpose(images, [0, 2, 1, 3, 4])
  images = np.reshape(images, [rows * width, cols * height, depth])
  return images


class SummaryWriter:
  """Saves data in event and summary protos for tensorboard."""

  def __init__(self, log_dir, enable=True):
    """Create a new SummaryWriter.

    Args:
      log_dir: path to record tfevents files in.
      enable: bool: if False don't actually write or flush data.  Used in
        multihost training.
    """
    # If needed, create log_dir directory as well as missing parent directories.
    if not tf.io.gfile.isdir(log_dir):
      tf.io.gfile.makedirs(log_dir)

    self._event_writer = EventFileWriter(log_dir, 10, 120, None)
    self._step = 0
    self._closed = False
    self._enabled = enable

  def add_summary(self, summary, step):
    if not self._enabled:
      return
    event = event_pb2.Event(summary=summary)
    event.wall_time = time.time()
    if step is not None:
      event.step = int(step)
    self._event_writer.add_event(event)

  def close(self):
    """Close SummaryWriter. Final!"""
    if not self._closed:
      self._event_writer.close()
      self._closed = True
      del self._event_writer

  def __del__(self):  # safe?
    # TODO(afrozm): Sometimes this complains with
    #  `TypeError: 'NoneType' object is not callable`
    try:
      self.close()
    except Exception:  # pylint: disable=broad-except
      pass

  def flush(self):
    if not self._enabled:
      return
    self._event_writer.flush()

  def scalar(self, tag, value, step=None):
    """Saves scalar value.

    Args:
      tag: str: label for this data
      value: int/float: number to log
      step: int: training step
    """
    value = float(np.array(value))
    if step is None:
      step = self._step
    else:
      self._step = step
    summary = tf.compat.v1.Summary(
        value=[tf.compat.v1.Summary.Value(tag=tag, simple_value=value)])
    self.add_summary(summary, step)

  def image(self, tag, image, step=None):
    """Saves RGB image summary from np.ndarray [H,W], [H,W,1], or [H,W,3].

    Args:
      tag: str: label for this data
      image: ndarray: [H,W], [H,W,1], [H,W,3] save image in greyscale or colors/
      step: int: training step
    """
    image = np.array(image)
    if step is None:
      step = self._step
    else:
      self._step = step
    if len(np.shape(image)) == 2:
      image = image[:, :, np.newaxis]
    if np.shape(image)[-1] == 1:
      image = np.repeat(image, 3, axis=-1)
    image_strio = io.BytesIO()
    plt.imsave(image_strio, image, format='png')
    image_summary = tf.compat.v1.Summary.Image(
        encoded_image_string=image_strio.getvalue(),
        colorspace=3,
        height=image.shape[0],
        width=image.shape[1])
    summary = tf.compat.v1.Summary(
        value=[tf.compat.v1.Summary.Value(tag=tag, image=image_summary)])
    self.add_summary(summary, step)

  def images(self, tag, images, step=None, rows=None, cols=None):
    """Saves (rows, cols) tiled images from np.ndarray.

    If either rows or cols aren't given, they are determined automatically
    from the size of the image batch, if neither are given a long column
    of images is produced. This truncates the image batch rather than padding
    if it doesn't fill the final row.

    Args:
      tag: str: label for this data
      images: ndarray: [N,H,W,1] or [N,H,W,3] to tile in 2d
      step: int: training step
      rows: int: number of rows in tile
      cols: int: number of columns in tile
    """
    images = np.array(images)
    if step is None:
      step = self._step
    else:
      self._step = step
    n_images = np.shape(images)[0]
    if rows is None and cols is None:
      rows = 1
      cols = n_images
    elif rows is None:
      rows = n_images // cols
    elif cols is None:
      cols = n_images // rows
    tiled_images = _pack_images(images, rows, cols)
    self.image(tag, tiled_images, step=step)

  def plot(self, tag, mpl_plt, step=None, close_plot=True):
    """Saves matplotlib plot output to summary image.

    Args:
      tag: str: label for this data
      mpl_plt: matplotlib stateful pyplot object with prepared plotting state
      step: int: training step
      close_plot: bool: automatically closes plot
    """
    if step is None:
      step = self._step
    else:
      self._step = step
    fig = mpl_plt.get_current_fig_manager()
    img_w, img_h = fig.canvas.get_width_height()
    image_buf = io.BytesIO()
    mpl_plt.savefig(image_buf, format='png')
    image_summary = tf.compat.v1.Summary.Image(
        encoded_image_string=image_buf.getvalue(),
        colorspace=4,  # RGBA
        height=img_h,
        width=img_w)
    summary = tf.compat.v1.Summary(
        value=[tf.compat.v1.Summary.Value(tag=tag, image=image_summary)])
    self.add_summary(summary, step)
    if close_plot:
      mpl_plt.close()

  def audio(self, tag, audiodata, step=None, sample_rate=44100):
    """Saves audio.

    NB: single channel only right now.

    Args:
      tag: str: label for this data
      audiodata: ndarray [Nsamples,]: data between (-1.0,1.0) to save as wave
      step: int: training step
      sample_rate: sample rate of passed in audio buffer
    """
    audiodata = np.array(audiodata)
    if step is None:
      step = self._step
    else:
      self._step = step
    audiodata = np.clip(np.squeeze(audiodata), -1, 1)
    if audiodata.ndim != 1:
      raise ValueError('Audio data must be 1D.')
    sample_list = (32767.0 * audiodata).astype(int).tolist()
    wio = io.BytesIO()
    wav_buf = wave.open(wio, 'wb')
    wav_buf.setnchannels(1)
    wav_buf.setsampwidth(2)
    wav_buf.setframerate(sample_rate)
    enc = b''.join([struct.pack('<h', v) for v in sample_list])
    wav_buf.writeframes(enc)
    wav_buf.close()
    encoded_audio_bytes = wio.getvalue()
    wio.close()
    audio = tf.compat.v1.Summary.Audio(
        sample_rate=sample_rate,
        num_channels=1,
        length_frames=len(sample_list),
        encoded_audio_string=encoded_audio_bytes,
        content_type='audio/wav')
    summary = tf.compat.v1.Summary(
        value=[tf.compat.v1.Summary.Value(tag=tag, audio=audio)])
    self.add_summary(summary, step)

  def histogram(self, tag, values, bins, step=None):
    """Saves histogram of values.

    Args:
      tag: str: label for this data
      values: ndarray: will be flattened by this routine
      bins: number of bins in histogram, or array of bins for np.histogram
      step: int: training step
    """
    if step is None:
      step = self._step
    else:
      self._step = step
    values = np.array(values)
    bins = np.array(bins)
    values = np.reshape(values, -1)
    counts, limits = np.histogram(values, bins=bins)
    # boundary logic
    cum_counts = np.cumsum(np.greater(counts, 0, dtype=np.int32))
    start, end = np.searchsorted(
        cum_counts, [0, cum_counts[-1] - 1], side='right')
    start, end = int(start), int(end) + 1
    counts = (
        counts[start -
               1:end] if start > 0 else np.concatenate([[0], counts[:end]]))
    limits = limits[start:end + 1]
    sum_sq = values.dot(values)
    histo = tf.compat.v1.HistogramProto(
        min=values.min(),
        max=values.max(),
        num=len(values),
        sum=values.sum(),
        sum_squares=sum_sq,
        bucket_limit=limits.tolist(),
        bucket=counts.tolist())
    summary = tf.compat.v1.Summary(
        value=[tf.compat.v1.Summary.Value(tag=tag, histo=histo)])
    self.add_summary(summary, step)

  def text(self, tag, textdata, step=None):
    """Saves a text summary.

    Args:
      tag: str: label for this data
      textdata: string, or 1D/2D list/numpy array of strings
      step: int: training step
    Note: markdown formatting is rendered by tensorboard.
    """
    if step is None:
      step = self._step
    else:
      self._step = step
    smd = tf.compat.v1.SummaryMetadata(
        plugin_data=tf.compat.v1.SummaryMetadata.PluginData(plugin_name='text'))
    if isinstance(textdata, (str, bytes)):
      tensor = tf.make_tensor_proto(
          values=[textdata.encode(encoding='utf_8')], shape=(1,))
    else:
      textdata = np.array(textdata)  # convert lists, jax arrays, etc.
      datashape = np.shape(textdata)
      if len(datashape) == 1:
        tensor = tf.make_tensor_proto(
            values=[td.encode(encoding='utf_8') for td in textdata],
            shape=(datashape[0],))
      elif len(datashape) == 2:
        tensor = tf.make_tensor_proto(
            values=[
                td.encode(encoding='utf_8') for td in np.reshape(textdata, -1)
            ],
            shape=(datashape[0], datashape[1]))
    summary = tf.compat.v1.Summary(
        value=[tf.compat.v1.Summary.Value(
            tag=tag, metadata=smd, tensor=tensor)])
    self.add_summary(summary, step)


# Copied from gin/tf/utils.py:GinConfigSaverHook
def markdownify_operative_config_str(string):
  """Convert an operative config string to markdown format."""

  # TODO(b/37527917): Total hack below. Implement more principled formatting.
  def process(line):
    """Convert a single line to markdown format."""
    if not line.startswith('#'):
      return '    ' + line

    line = line[2:]
    if line.startswith('===='):
      return ''
    if line.startswith('None'):
      return '    # None.'
    if line.endswith(':'):
      return '#### ' + line
    return line

  output_lines = []
  for line in string.splitlines():
    procd_line = process(line)
    if procd_line is not None:
      output_lines.append(procd_line)

  return '\n'.join(output_lines)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for trax.data.inputs."""

import itertools
import os

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
from trax import data

pkg_dir, _ = os.path.split(__file__)
_TESTDATA = os.path.join(pkg_dir, 'testdata')


def _spm_path():
  return os.path.join(_TESTDATA, 'sentencepiece.model')


class InputsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('zero', 0),
      ('negative', -5),
  )
  def test_shuffle_data_raises_error_queue_size(self, queue_size):
    samples = iter(range(10))
    with self.assertRaises(ValueError):
      _ = list(data.shuffle(samples, queue_size))

  @parameterized.named_parameters(
      ('one', 1),
      ('two', 2),
      ('twenty', 20),
  )
  def test_shuffle_data_queue_size(self, queue_size):
    samples = iter(range(100, 200))
    shuffled_stream = data.shuffle(samples, queue_size)
    first_ten = [next(shuffled_stream) for _ in range(10)]

    # Queue size limits how far ahead/upstream the current sample can reach.
    self.assertLess(first_ten[0], 100 + queue_size)
    self.assertLess(first_ten[3], 103 + queue_size)
    self.assertLess(first_ten[9], 109 + queue_size)

    unshuffled_first_ten = list(range(100, 110))
    if queue_size == 1:  # Degenerate case: no shuffling can happen.
      self.assertEqual(first_ten, unshuffled_first_ten)
    if queue_size > 1:
      self.assertNotEqual(first_ten, unshuffled_first_ten)

  @parameterized.named_parameters(
      ('qsize_100_n_001', 100, 1),
      ('qsize_100_n_099', 100, 99),
      ('qsize_100_n_100', 100, 100),
      ('qsize_100_n_101', 100, 101),
      ('qsize_100_n_199', 100, 199),
  )
  def test_shuffle_data_yields_all_samples(self, queue_size, n_samples):
    samples = iter(range(n_samples))
    shuffled_stream = data.shuffle(samples, queue_size)
    self.assertLen(list(shuffled_stream), n_samples)

  def test_batch_data(self):
    dataset = ((i, i+1) for i in range(10))
    batches = data.batch(dataset, 10)
    batch = next(batches)
    self.assertLen(batch, 2)
    self.assertEqual(batch[0].shape, (10,))

  def test_batch_data_padding(self):
    dataset = (([1] * (10 - i), i+1) for i in range(10))
    batches = data.batch(dataset, 10)
    batch = next(batches)
    self.assertEqual(batch[0].shape, (10, 10))
    self.assertTrue(np.array_equal(batch[0][-1], np.asarray([1] + 9 * [0])))

  def test_batch_exception_size(self):
    dataset = ((i, i + 1) for i in range(10))
    with self.assertRaises(ValueError):
      batches = data.batch(dataset, 0)
      next(batches)

  def test_serial(self):
    dataset = lambda _: ((i, i+1) for i in range(10))
    batches = data.Serial(dataset, data.Shuffle(3), data.Batch(10))
    batch = next(batches())
    self.assertLen(batch, 2)
    self.assertEqual(batch[0].shape, (10,))

  def test_serial_composes(self):
    """Check that data.Serial works inside another data.Serial."""
    dataset = lambda _: ((i, i+1) for i in range(10))
    serial1 = data.Serial(dataset, data.Shuffle(3))
    batches = data.Serial(serial1, data.Batch(10))
    batch = next(batches())
    self.assertLen(batch, 2)
    self.assertEqual(batch[0].shape, (10,))

  def test_count_and_skip(self):
    dataset = lambda _: ((i, i+1) for i in range(10))
    examples = data.Serial(dataset, data.CountAndSkip('toy_data'))
    ex_generator = examples()
    ex1 = next(ex_generator)
    self.assertEqual(ex1, (0, 1))
    self.assertEqual(data.inputs.data_counters['toy_data'], 1)
    ex2 = next(ex_generator)
    self.assertEqual(ex2, (1, 2))
    self.assertEqual(data.inputs.data_counters['toy_data'], 2)
    ex3 = next(examples())  # new generator, will skip
    self.assertEqual(ex3, (2, 3))
    self.assertEqual(data.inputs.data_counters['toy_data'], 3)
    data.inputs.data_counters['toy_data'] = 0  # reset
    ex4 = next(examples())  # new generator, was reset
    self.assertEqual(ex4, (0, 1))
    self.assertEqual(data.inputs.data_counters['toy_data'], 1)

  def test_parallel(self):
    """Basic test of the parallel ccmbinator."""
    dataset1 = lambda: (i for i in range(10))
    dataset2 = lambda: (i for i in range(10, 20))
    parallel = data.Parallel([dataset1, dataset2])
    generator = parallel()

    self.assertEqual(next(generator), 0)
    self.assertEqual(next(generator), 10)
    self.assertEqual(next(generator), 1)
    self.assertEqual(next(generator), 11)
    self.assertEqual(next(generator), 2)
    self.assertEqual(next(generator), 12)

  def test_parallel_with_gen_not_none(self):
    """Test of the parallel ccmbinator with a not none generator."""
    dataset1 = lambda _: (i for i in range(10))
    dataset2 = lambda _: (i for i in range(10, 20))
    parallel = data.Parallel([dataset1, dataset2])

    def test_generator():
      yield 0

    generator = parallel(gen=test_generator)

    self.assertEqual(next(generator), 0)
    self.assertEqual(next(generator), 10)
    self.assertEqual(next(generator), 1)
    self.assertEqual(next(generator), 11)
    self.assertEqual(next(generator), 2)
    self.assertEqual(next(generator), 12)

  def test_parallel_with_weights(self):
    """Test of the parallel ccmbinator with weights."""
    dataset1 = lambda: (i for i in range(10))
    dataset2 = lambda: (i for i in range(10, 20))
    parallel = data.Parallel([dataset1, dataset2], counters=(2, 1))
    generator = parallel()

    self.assertEqual(next(generator), 0)
    self.assertEqual(next(generator), 10)
    self.assertEqual(next(generator), 1)
    self.assertEqual(next(generator), 11)
    self.assertEqual(next(generator), 2)
    self.assertEqual(next(generator), 3)
    self.assertEqual(next(generator), 12)
    self.assertEqual(next(generator), 4)
    self.assertEqual(next(generator), 5)
    self.assertEqual(next(generator), 13)

  def test_parallel_with_weights_and_minimum(self):
    """Test of the parallel ccmbinator with weights and minimum."""
    dataset1 = lambda: (i for i in range(10))
    dataset2 = lambda: (i for i in range(10, 110))
    parallel = data.Parallel([dataset1, dataset2],
                             counters=(10, 100),
                             reweight_by_minimum=True)
    generator = parallel()

    self.assertEqual(next(generator), 0)
    self.assertEqual(next(generator), 10)
    self.assertEqual(next(generator), 11)
    self.assertEqual(next(generator), 12)
    self.assertEqual(next(generator), 13)
    self.assertEqual(next(generator), 14)
    self.assertEqual(next(generator), 15)
    self.assertEqual(next(generator), 16)
    self.assertEqual(next(generator), 17)
    self.assertEqual(next(generator), 18)
    self.assertEqual(next(generator), 19)
    self.assertEqual(next(generator), 1)
    self.assertEqual(next(generator), 20)
    self.assertEqual(next(generator), 21)
    self.assertEqual(next(generator), 22)
    self.assertEqual(next(generator), 23)
    self.assertEqual(next(generator), 24)
    self.assertEqual(next(generator), 25)
    self.assertEqual(next(generator), 26)
    self.assertEqual(next(generator), 27)
    self.assertEqual(next(generator), 28)
    self.assertEqual(next(generator), 29)
    self.assertEqual(next(generator), 2)

  def test_parallel_with_gradual_reweighting(self):
    """Test of the parallel ccmbinator with weights and minimum."""
    dataset1 = lambda: (i for i in itertools.cycle(range(1)))
    dataset2 = lambda: (i for i in itertools.cycle(range(10, 30)))
    dataset3 = lambda: (i for i in itertools.cycle(range(30, 70)))
    parallel = data.Parallel([dataset2, dataset1, dataset3],
                             counters=(20, 1, 40),
                             gradually_reweight=True)
    generator = parallel()

    for _ in range(3):
      self.assertEqual(next(generator), 0)
      for i in range(20):
        self.assertEqual(next(generator), 10 + i)
        self.assertEqual(next(generator), 30 + 2 * i)
        self.assertEqual(next(generator), 30 + 2 * i + 1)

  def test_parallel_with_gradual_reweighting_remainders(self):
    """Test of the parallel ccmbinator with weights and minimum."""
    dataset1 = lambda: (i for i in itertools.cycle(range(1)))
    dataset2 = lambda: (i for i in itertools.cycle(range(10, 30)))
    dataset3 = lambda: (i for i in itertools.cycle(range(30, 80)))
    parallel = data.Parallel([dataset2, dataset1, dataset3],
                             counters=(20, 1, 50),
                             gradually_reweight=True,
                             use_remainders=True)
    generator = parallel()

    for _ in range(3):
      self.assertEqual(next(generator), 0)
      for i in range(20):
        self.assertEqual(next(generator), 10 + i)
        self.assertEqual(next(generator), 30 + 2 * i)
        self.assertEqual(next(generator), 30 + 2 * i + 1)
      # Here we process the remainder from dataset 3:
      for i in range(10):
        self.assertEqual(next(generator), 70 + i)

  def test_parallel_with_gradual_reweighting_remainders_big(self):
    """Test of the parallel ccmbinator with weights and minimum."""
    dataset1 = lambda: (i for i in itertools.cycle(range(1)))
    dataset2 = lambda: (i for i in itertools.cycle(range(10, 30)))
    dataset3 = lambda: (i for i in itertools.cycle(range(30, 80)))
    dataset4 = lambda: (i for i in itertools.cycle(range(100, 220)))
    parallel = data.Parallel([dataset2, dataset1, dataset4, dataset3],
                             counters=(20, 1, 120, 50),
                             gradually_reweight=True,
                             use_remainders=True)
    generator = parallel()

    for _ in range(3):
      self.assertEqual(next(generator), 0)
      for i in range(20):
        self.assertEqual(next(generator), 10 + i)
        for j in range(2):
          self.assertEqual(next(generator), 30 + 2 * i + j)
          for k in range(2):
            self.assertEqual(next(generator), 100 + 2 * 2 * i + 2 * j + k)
      # Here we process the remainder from datasets 3 and 4:
      for i in range(10):
        self.assertEqual(next(generator), 70 + i)
      for i in range(40):
        self.assertEqual(next(generator), 180 + i)

  def test_parallel_with_weights_three_datasets(self):
    """Check that data.Serial works inside another data.Serial."""
    dataset1 = lambda: (i for i in range(10))
    dataset2 = lambda: (i for i in range(10, 20))
    dataset3 = lambda: (i for i in range(20, 30))
    parallel = data.Parallel(
        [dataset1, dataset2, dataset3], counters=(2, 1, 3))
    generator = parallel()

    self.assertEqual(next(generator), 0)    # (1,0,0)
    self.assertEqual(next(generator), 10)   # (1,1,0)
    self.assertEqual(next(generator), 20)   # (1,1,1)
    self.assertEqual(next(generator), 1)    # (2,1,1)
    self.assertEqual(next(generator), 21)   # (2,1,2)
    self.assertEqual(next(generator), 22)   # (2,1,3)
    self.assertEqual(next(generator), 2)    # (1,0,0)
    self.assertEqual(next(generator), 11)   # (1,1,0)
    self.assertEqual(next(generator), 23)   # (1,1,1)
    self.assertEqual(next(generator), 3)    # (2,1,1)
    self.assertEqual(next(generator), 24)   # (2,1,2)
    self.assertEqual(next(generator), 25)   # (2,1,3)
    self.assertEqual(next(generator), 4)    # (1,0,0)

  def test_stack_parallel(self):
    """Test of stacked parallel ccmbinators."""
    dataset1 = lambda: (i for i in range(10))
    dataset2 = lambda: (i for i in range(10, 20))
    dataset3 = lambda: (i for i in range(20, 30))
    parallel_lev0 = data.Parallel([dataset1, dataset2])
    parallel_lev1 = data.Parallel([parallel_lev0, dataset3])
    generator = parallel_lev1()

    self.assertEqual(next(generator), 0)
    self.assertEqual(next(generator), 20)
    self.assertEqual(next(generator), 10)
    self.assertEqual(next(generator), 21)
    self.assertEqual(next(generator), 1)
    self.assertEqual(next(generator), 22)
    self.assertEqual(next(generator), 11)
    self.assertEqual(next(generator), 23)
    self.assertEqual(next(generator), 2)
    self.assertEqual(next(generator), 24)
    self.assertEqual(next(generator), 12)

  def test_parallel_with_zero_counters(self):
    """Test of stacked parallel ccmbinators."""
    dataset1 = lambda: (i for i in range(10))
    dataset2 = lambda: (i for i in range(10, 20))
    dataset3 = lambda: (i for i in range(20, 30))
    parallel = data.Parallel([dataset1, dataset2, dataset3], counters=[1, 0, 1])
    generator = parallel()

    self.assertEqual(next(generator), 0)
    self.assertEqual(next(generator), 20)
    self.assertEqual(next(generator), 1)
    self.assertEqual(next(generator), 21)
    self.assertEqual(next(generator), 2)
    self.assertEqual(next(generator), 22)
    self.assertEqual(next(generator), 3)
    self.assertEqual(next(generator), 23)

  def test_serial_with_python(self):
    dataset = lambda _: ((i, i+1) for i in range(10))
    batches = data.Serial(
        dataset,
        lambda g: map(lambda x: (x[0], x[1] + 1), g),
        lambda g: filter(lambda x: x[0] % 2 == 1, g),
        data.Batch(2)
    )
    batch = next(batches())
    self.assertLen(batch, 2)
    (xs, ys) = batch
    # First tuple after filtering is (1, 3) = (1, 2+1).
    self.assertEqual(xs[0], 1)
    self.assertEqual(ys[0], 3)
    # Second tuple after filtering is (3, 5).
    self.assertEqual(xs[1], 3)
    self.assertEqual(ys[1], 5)

  def test_pad_to_max_dims(self):
    tensors1 = [np.zeros((3, 10)), np.ones((3, 10))]
    padded1 = data.inputs.pad_to_max_dims(tensors1)
    self.assertEqual(padded1.shape, (2, 3, 10))
    tensors2 = [np.zeros((2, 10)), np.ones((3, 9))]
    padded2 = data.inputs.pad_to_max_dims(tensors2)
    self.assertEqual(padded2.shape, (2, 3, 10))
    tensors3 = [np.zeros((8, 10)), np.ones((8, 9))]
    padded3 = data.inputs.pad_to_max_dims(tensors3, 12)
    self.assertEqual(padded3.shape, (2, 12, 12))
    tensors4 = [np.zeros((2, 10)), np.ones((3, 9))]
    padded4 = data.inputs.pad_to_max_dims(tensors4, 12)
    self.assertEqual(padded4.shape, (2, 4, 12))

  def test_pad_to_length(self):
    tensors1 = [(np.zeros((5)), np.ones((3)))]
    pad_to_length_function1 = data.inputs.PadToLength(len_map={0: 10,
                                                               1: 11},
                                                      pad_value={0: 0,
                                                                 1: 1})
    padded1 = next(pad_to_length_function1(tensors1))
    self.assertEqual(padded1[0].shape, (10,))
    self.assertEqual(padded1[1].shape, (11,))

    tensors2 = [(np.zeros((15)), np.ones((20)))]
    pad_to_length_function2 = data.inputs.PadToLength(len_map={0: 10,
                                                               1: 10},
                                                      pad_value={0: 0,
                                                                 1: 1},
                                                      multiple=True)
    padded2 = next(pad_to_length_function2(tensors2))
    self.assertEqual(padded2[0].shape, (20,))
    self.assertEqual(padded2[1].shape, (20,))

  def test_concatenate_lm_input(self):
    tensors1 = [(np.zeros((5)), np.ones((3)))]

    lm_input_function1 = data.inputs.ConcatenateToLMInput(pad_to_length=10)
    lm_input_1 = next(lm_input_function1(tensors1))
    self.assertEqual(lm_input_1[0].shape, (10,))
    self.assertEqual(lm_input_1[1].shape, (10,))
    self.assertEqual(lm_input_1[2].shape, (10,))
    self.assertEqual(lm_input_1[2].all(),
                     np.array([[0., 0., 0., 0., 0.,
                                1., 1., 1., 0., 0.]]).all())

    tensors2 = [(np.zeros((5)), np.ones((3)))]
    lm_input_function2 = data.inputs.ConcatenateToLMInput()
    lm_input_2 = next(lm_input_function2(tensors2))
    self.assertEqual(lm_input_2[0].shape, (8,))
    self.assertEqual(lm_input_2[1].shape, (8,))
    self.assertEqual(lm_input_2[2].shape, (8,))
    self.assertEqual(lm_input_2[2].all(),
                     np.array([[0., 0., 0., 0., 0.,
                                1., 1., 1.]]).all())

  def test_truncate_to_length_no_arg(self):
    """Tests that a no-arg call leaves shapes unchanged."""
    def data_stream():
      while True:
        yield (np.zeros((1, 5)), np.ones((1, 5)))
    stream_fn = data.inputs.TruncateToLength()
    y0, y1 = next(stream_fn(data_stream()))
    self.assertEqual(y0.shape, (1, 5))
    self.assertEqual(y1.shape, (1, 5))

  @parameterized.named_parameters(
      ('none', None, ((1, 5), (1, 5))),
      ('large_values', {0: (1, 77), 1: (1, 88)}, ((1, 5), (1, 5))),
      ('small_values', {0: (1, 3), 1: (1, 2)}, ((1, 3), (1, 2))),
  )
  def test_truncate_to_length_len_map(self, len_map, out_shapes):
    """Tests that truncation occurs when len_map values are small enough."""
    def data_stream():
      while True:
        yield (np.zeros((1, 5)), np.ones((1, 5)))
    stream_fn = data.inputs.TruncateToLength(len_map=len_map)
    y0, y1 = next(stream_fn(data_stream()))
    self.assertEqual(y0.shape, out_shapes[0])
    self.assertEqual(y1.shape, out_shapes[1])

  def test_truncate_to_length_questionable_behavior(self):
    # Use of np.reshape in TruncateToLength allows non-truncation results
    # without warning. As long as the target shape (len_map value) is
    # lexicographically prior to the data shape, then np.reshape can happen,
    # even if it results in *adding* values to the overall array.
    #
    # This test passes as a marker of the questionable behavior, and should
    # *fail* -- and then be removed -- when the function is
    # clarified/re-implemented.
    #
    # TODO(jonni): Determine desired behavior, and fit implementation to it.
    x = np.arange(21).reshape((1, 21, 1))
    def data_stream():
      while True:
        yield x
    stream_fn = data.inputs.TruncateToLength(len_map={0: (1, 4, 6)})
    (y,) = next(stream_fn(data_stream()))
    self.assertEqual(y.shape, (1, 4, 6))
    self.assertEqual(y[0, 3, 1], 19)
    self.assertEqual(y[0, 3, 2], 20)  # end of original values [0..20]
    self.assertEqual(y[0, 3, 3], 0)  # added value
    self.assertEqual(y[0, 3, 4], 1)  # added value
    self.assertEqual(y[0, 3, 5], 2)  # added value

  def test_filter_empty_examples(self):
    tensors1 = [(np.zeros((0,)), np.ones((1, 5))),
                (np.zeros((1, 5)), np.ones((1, 5)))]

    filter_empty_examples_function1 = data.inputs.FilterEmptyExamples()
    filtered1 = next(filter_empty_examples_function1(tensors1))
    self.assertEqual(filtered1[0].shape, (1, 5))
    self.assertEqual(filtered1[1].shape, (1, 5))

    filter_empty_examples_function2 = data.inputs.FilterEmptyExamples(axes=[1])
    filtered2 = next(filter_empty_examples_function2(tensors1))
    self.assertEqual(filtered2[0].shape, (0,))
    self.assertEqual(filtered2[1].shape, (1, 5))

  def test_append_value(self):
    tensors1 = [(np.zeros((1, 5)), np.ones((1, 5)))]

    append_value_function1 = data.inputs.AppendValue()
    unmodified = next(append_value_function1(tensors1))
    self.assertEqual(unmodified[0].shape, (1, 5))
    self.assertEqual(unmodified[1].shape, (1, 5))

    append_value_function2 = data.inputs.AppendValue({0: [[5]],
                                                      1: [[4]]})
    appended = next(append_value_function2(tensors1))
    self.assertEqual(appended[0].shape, (1, 6))
    self.assertEqual(appended[0].all(),
                     np.array([[0., 0., 0., 0., 0., 5.]]).all())
    self.assertEqual(appended[1].shape, (1, 6))
    self.assertEqual(appended[1].all(),
                     np.array([[1., 1., 1., 1., 1., 4.]]).all())

  def test_pad_to_max_dims_boundary_list(self):
    tensors = [np.zeros((1, 15, 31)), np.ones((2, 10, 35)), np.ones((4, 2, 3))]
    padded_tensors = data.inputs.pad_to_max_dims(
        tensors, boundary=(None, 15, 20))
    # no boundary, only max in the first dim, 15 is already the max len in
    # second dim, last dim padded to multiple of 20.
    # The outer dim is the batch here.
    self.assertEqual(padded_tensors.shape, (3, 4, 15, 40))

  def test_pad_to_max_dims_strict_pad_on_len(self):
    tensors = [np.ones((15,)), np.ones((12,)), np.ones((14,))]
    padded_tensors = data.inputs.pad_to_max_dims(
        tensors, boundary=10, strict_pad_on_len=True)
    self.assertEqual(padded_tensors.shape, (3, 20))

  def test_bucket_by_length(self):
    def fake_generator(length, num_examples=1):
      for _ in range(num_examples):
        yield (np.ones((length,)), np.ones((length,)))

    def length_function(example):
      return max(example[0].shape[0], example[1].shape[0])

    batches = list(data.bucket_by_length(fake_generator(5, 6),
                                         length_function,
                                         [20],
                                         [2],
                                         strict_pad_on_len=True))

    # We'll get three batches of 2 examples each.
    self.assertLen(batches, 3)
    self.assertIsInstance(batches[0], tuple)
    self.assertLen(batches[0], 2)
    self.assertEqual((2, 20), batches[0][0].shape)
    self.assertEqual((2, 20), batches[0][1].shape)

  @parameterized.named_parameters(
      ('encdec_on', True),
      ('encdec_off', False),
  )
  def test_addition_inputs_exceptions(self, encdec):
    vocab_size = 5
    batch_size = 256
    seq_length = 64
    # Check if max/min lengths are validated for train stream
    with self.assertRaises(ValueError):
      inputs = data.inputs.addition_inputs(
          vocab_size=vocab_size,
          batch_size=batch_size,
          train_length=2,
          eval_min_length=1,
          eval_max_length=seq_length,
          pad_to_multiple=seq_length,
          encdec=encdec)
      train_stream = inputs.train_stream(n_devices=1)
      for _ in range(10):
        next(train_stream)

    # Check if max/min lengths are validated for eval stream
    with self.assertRaises(ValueError):
      inputs = data.inputs.addition_inputs(
          vocab_size=vocab_size,
          batch_size=batch_size,
          train_length=seq_length,
          eval_min_length=1,
          eval_max_length=seq_length,
          pad_to_multiple=seq_length,
          encdec=True)
      eval_stream = inputs.eval_stream(n_devices=1)
      for _ in range(10):
        next(eval_stream)

  def test_addition_inputs_constraints(self):
    vocab_size = 5
    batch_size = 256
    seq_length = 64
    inputs = data.inputs.addition_inputs(
        vocab_size=vocab_size,
        batch_size=batch_size,
        train_length=seq_length,
        eval_min_length=seq_length,
        eval_max_length=seq_length,
        pad_to_multiple=seq_length,
        encdec=True)

    # Check if max length is respected for train stream
    train_stream = inputs.train_stream(n_devices=1)
    for _ in range(10):
      x, y, weights = next(train_stream)
      self.assertEqual(x.shape[1], seq_length)
      self.assertEqual(y.shape[1], seq_length)
      self.assertEqual(weights.shape[1], seq_length)

    # Check if max length is respected for eval stream
    eval_stream = inputs.eval_stream(n_devices=1)
    for _ in range(10):
      x, y, weights = next(eval_stream)
      self.assertEqual(x.shape[1], seq_length)
      self.assertEqual(y.shape[1], seq_length)
      self.assertEqual(weights.shape[1], seq_length)

  def _get_span_lengths(self, x):
    span_lengths = []
    curr_len = 0
    for i in range(1, len(x)):
      # 1 -> 0
      if x[i] == 0 and x[i - 1] == 1:
        span_lengths.append(curr_len)
        curr_len = 0
      # 1 -> 1 or 0 -> 1
      elif ((x[i] == 1 and x[i - 1] == 1) or
            (x[i] == 1 and x[i - 1] == 0)):
        curr_len += 1
    if curr_len != 0:
      span_lengths.append(curr_len)
    return span_lengths

  def test_random_spans_noise_mask(self):
    length = 100
    noise_density = 0.15
    mean_noise_span_length = 3.0

    # Take 5 random seed1, seed2 values.
    for seed in np.random.randint(0, 100, (5, 2)):
      is_noise = data.random_spans_noise_mask(length,
                                              noise_density,
                                              mean_noise_span_length,
                                              seed1=seed[0],
                                              seed2=seed[1])
      is_noise = is_noise.astype(np.int32)
      # noise_density fraction of tokens are produced
      self.assertEqual(np.sum(is_noise), noise_density * length)
      # Get span lengths and make sure the average is what we expect.
      actual_span_lengths = self._get_span_lengths(is_noise)
      average_span_length = (
          sum(actual_span_lengths) / len(actual_span_lengths))
      self.assertEqual(mean_noise_span_length, average_span_length)

  def test_process_c4_with_span_corruption(self):
    def process_c4_with_span_corruption(spm_path=None,
                                        extra_ids=0,
                                        train=False,
                                        max_length=100,
                                        noise_density=0.15,
                                        mean_noise_span_length=3.0,
                                        seed1=None,
                                        seed2=None):
      return data.Serial(
          data.TFDS(
              'c4/en:2.3.0', data_dir=_TESTDATA, keys=('text',), train=train),
          data.SentencePieceTokenize(spm_path=spm_path, extra_ids=extra_ids),
          data.generate_sequential_chunks(max_length=max_length),
          data.generate_random_noise_mask(
              noise_density=noise_density,
              mean_noise_span_length=mean_noise_span_length,
              seed1=seed1, seed2=seed2),
          data.consume_noise_mask(vocab_size=32000 + extra_ids),
          data.FilterEmptyExamples(),
          data.AppendValue(val={0: [1], 1: [1]}),
          data.PadToLength(len_map={0: 100, 1: 30}, pad_value={0: 0, 1: 0}),
          data.AddLossWeights(id_to_mask=0),
          data.Batch(batch_size=2)
          )

    gen = process_c4_with_span_corruption(
        spm_path=_spm_path(), seed1=0, seed2=1)

    examples = []
    for i, ex in enumerate(gen()):
      if i == 100:
        break
      examples.append(ex)

    self.assertLen(examples, 100)
    example = examples[0]

    batched_input, batched_output, batched_loss_weights = example

    self.assertSequenceEqual(
        batched_input.tolist(),
        # pylint: disable=bad-continuation,bad-whitespace
        [[   37,  2335,   113,  3977,   227,  7306,    45,     3,     9,
           4716,   147,     8,    71,  2658,    65,   118,  4313,    38,
              3,     9, 13065,    32, 31999,     9,  5704,    26,   109,
              6,  6862,     6,  4728,    45,     8,  3796, 24093, 11834,
           4716,    30,     8,  1379,    13, 31998,   130,   718,    12,
              8, 24124,  1343,   300,  4357,  1714, 31997,  1373,    47,
          16487,  3168,    16,   321,  7943,     5,     3,  4868,  3856,
           5700,    75,     7,   200,  2231,     6, 11163,     9,     6,
            113,    47,  5330,    45, 14354,     6,    47, 31996, 20721,
           3654,    44,     8,  3112,     5, 14599,    11,  8067, 31995,
              1,     0,     0,     0,     0,     0,     0,     0,     0,
              0],
         [  277,   828,    43,  5899,    46,    16, 10952,   139,   160,
           1687,    56,   539,    30,  2875,    41, 31122,  2307,   137,
           2702,  2780,    15,     7, 31999,    44,     8,  3112,    11,
             30,   569,   783,     5,     3, 17701,     6,  2194,    26,
             23,  1336,  6321,  1694,    30, 31998,   196,    56,  1852,
           1423,    25,     5,    27,   183,  8032, 31997,   217,   149,
           1513,    11,  2238,    25,  1800,     5,    96,  2703,    44,
           3065, 12537, 11163,     9,   535,    71,  9363, 14886,   646,
             44,     8,  3112,   243, 23281,    12,     8, 31996,   346,
            402,    17,    99,    83,    11,   773,  3668,  1280, 31995,
              1,     0,     0,     0,     0,     0,     0,     0,     0,
              0]]
        # pylint: enable=bad-continuation,bad-whitespace
        )

    self.assertSequenceEqual(
        batched_output.tolist(),
        # pylint: disable=bad-continuation,bad-whitespace
        [[31999,  1639,     7, 15480,     5, 11163, 31998,  2083,  9997,
           5076, 31997,   265,    11,     8, 31996,     3, 31995,  1343,
           2487,   106,     1,     0,     0,     0,     0,     0,     0,
              0,     0,     0],
         [31999,    12,     8, 15480,   130,   646, 31998,  1376,    10,
             96, 31997,    62,   410,    59, 31996,    96, 31995,    94,
            608,    10,     1,     0,     0,     0,     0,     0,     0,
              0,     0,     0]]
        # pylint: enable=bad-continuation,bad-whitespace
        )

    self.assertSequenceEqual(
        batched_loss_weights.tolist(),
        # pylint: disable=bad-continuation,bad-whitespace
        [[1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
          1., 1., 1., 1., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
         [1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
          1., 1., 1., 1., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0.]]
        # pylint: enable=bad-continuation,bad-whitespace
        )

  def test_prefix_lm_last_output_batch_is_short(self):
    prefix_lm_fn = data.PrefixLM(input_length=2, output_length=3)
    examples = list(prefix_lm_fn([[1, 2, 3, 4, 5, 6, 7, 8]]))
    self.assertSequenceEqual(([1, 2], [3, 4, 5]), examples[0])
    self.assertSequenceEqual(([6, 7], [8]), examples[1])
    self.assertLen(examples, 2)

  def test_prefix_lm_last_input_batch_is_short(self):
    prefix_lm_fn = data.PrefixLM(input_length=2, output_length=3)
    examples = list(prefix_lm_fn([[1, 2, 3, 4, 5, 6]]))
    self.assertSequenceEqual(([1, 2], [3, 4, 5]), examples[0])
    self.assertLen(examples, 1)

  def test_prefix_lm_last_input_batch_exists_but_no_output(self):
    prefix_lm_fn = data.PrefixLM(input_length=2, output_length=3)
    examples = list(prefix_lm_fn([[1, 2, 3, 4, 5, 6, 7]]))
    self.assertSequenceEqual(([1, 2], [3, 4, 5]), examples[0])
    self.assertLen(examples, 1)

  def test_unbatch(self):
    unbatch_fn = data.UnBatch()
    batched_inputs = [
        # First batch - 3 examples
        (np.arange(3*2).reshape(3, -1),
         np.arange(3*3).reshape(3, -1),
         np.arange(3*4).reshape(3, -1)),
        # Second batch - 4 examples
        (np.arange(4*2).reshape(4, -1),
         np.arange(4*3).reshape(4, -1),
         np.arange(4*4).reshape(4, -1)),
    ]
    examples = list(unbatch_fn(batched_inputs))
    self.assertLen(examples, 3 + 4)

  def test_sine_shape(self):
    inputs = data.sine_inputs(batch_size=3, length=5)
    train_batch = next(inputs.train_stream(n_devices=1))
    eval_batch = next(inputs.eval_stream(n_devices=1))
    # (observations, actions, observations, mask)
    self.assertLen(train_batch, 4)
    self.assertLen(eval_batch, 4)
    for (x, y) in zip(train_batch, eval_batch):
      self.assertEqual(x.shape, (3, 5))
      self.assertEqual(y.shape, (3, 5))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Models defined in trax."""
import gin

from trax.models import atari_cnn
from trax.models import mlp
from trax.models import neural_gpu
from trax.models import resnet
from trax.models import rl
from trax.models import rnn
from trax.models import transformer
from trax.models.reformer import reformer
from trax.models.research import bert
from trax.models.research import configurable_transformer
from trax.models.research import hourglass
from trax.models.research import layerdrop_transformer
from trax.models.research import rezero
from trax.models.research import rse
from trax.models.research import terraformer
from trax.models.research import transformer2


# Ginify
def model_configure(*args, **kwargs):
  kwargs['module'] = 'trax.models'
  return gin.external_configurable(*args, **kwargs)


# pylint: disable=invalid-name
AtariCnn = model_configure(atari_cnn.AtariCnn)
AtariCnnBody = model_configure(atari_cnn.AtariCnnBody)
FrameStackMLP = model_configure(atari_cnn.FrameStackMLP)
BERT = model_configure(bert.BERT)
BERTClassifierHead = model_configure(bert.BERTClassifierHead)
BERTRegressionHead = model_configure(bert.BERTRegressionHead)
ConfigurableTerraformer = model_configure(terraformer.ConfigurableTerraformer)
ConfigurableTransformer = model_configure(
    configurable_transformer.ConfigurableTransformer)
ConfigurableTransformerEncoder = model_configure(
    configurable_transformer.ConfigurableTransformerEncoder)
ConfigurableTransformerLM = model_configure(
    configurable_transformer.ConfigurableTransformerLM)
MLP = model_configure(mlp.MLP)
NeuralGPU = model_configure(neural_gpu.NeuralGPU)
Reformer = model_configure(reformer.Reformer)
ReformerLM = model_configure(reformer.ReformerLM)
ReformerShortenLM = model_configure(reformer.ReformerShortenLM)
Resnet50 = model_configure(resnet.Resnet50)
ReZeroTransformer = model_configure(
    rezero.ReZeroTransformer)
ReZeroTransformerDecoder = model_configure(
    rezero.ReZeroTransformerDecoder)
ReZeroTransformerEncoder = model_configure(
    rezero.ReZeroTransformerEncoder)
ReZeroTransformerLM = model_configure(
    rezero.ReZeroTransformerLM)
SkippingTransformerLM = model_configure(
    layerdrop_transformer.SkippingTransformerLM)
LayerDropTransformerLM = model_configure(
    layerdrop_transformer.LayerDropTransformerLM)
EveryOtherLayerDropTransformerLM = model_configure(
    layerdrop_transformer.EveryOtherLayerDropTransformerLM)
Transformer = model_configure(transformer.Transformer)
TransformerDecoder = model_configure(transformer.TransformerDecoder)
TransformerEncoder = model_configure(transformer.TransformerEncoder)
TransformerLM = model_configure(transformer.TransformerLM)
Transformer2 = model_configure(
    transformer2.Transformer2)
WideResnet = model_configure(resnet.WideResnet)
Policy = model_configure(rl.Policy)
PolicyAndValue = model_configure(rl.PolicyAndValue)
Value = model_configure(rl.Value)
Quality = model_configure(rl.Quality)
RNNLM = model_configure(rnn.RNNLM)
GRULM = model_configure(rnn.GRULM)
LSTMSeq2SeqAttn = model_configure(rnn.LSTMSeq2SeqAttn)
ResidualShuffleExchange = model_configure(rse.ResidualShuffleExchange)
HourglassLM = model_configure(hourglass.HourglassLM)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Experimenting with position encodings."""

import logging
import jax
import numpy as np
import trax
from trax import fastmath
from trax.fastmath import numpy as jnp
from trax.layers import base as layer_base
from trax.layers import initializers as init


class AxialPositionalEncoding(layer_base.Layer):
  """Axial positional encoding."""
  # TODO(kitaev): support variable-length sequences.

  def __init__(self, shape=(64, 64, 3), d_embs=(384, 384, 256),
               kernel_initializer=init.RandomNormalInitializer(1.0),
               dropout=0.0, dropout_broadcast_dims=(), mode='train'):
    super().__init__()
    self._kernel_initializer = kernel_initializer
    assert len(shape) == len(d_embs)
    self._shape = shape
    self._d_embs = d_embs

    if dropout >= 1.0:
      raise ValueError('Dropout rates must be lower than 1.')
    if mode == 'train':
      self._dropout = dropout
    else:
      self._dropout = 0.0
    self._dropout_broadcast_dims = dropout_broadcast_dims
    self._mode = mode

  def forward(self, inputs):
    rng, state = self.rng, self.state
    embs = []
    for ax_emb in self.weights:
      ax_emb = jnp.broadcast_to(
          ax_emb, (inputs.shape[0],) + self._shape + (ax_emb.shape[-1],))
      embs.append(ax_emb)

    if self._mode == 'predict':
      assert self._dropout == 0.0
      emb = jnp.concatenate(embs, -1)
      emb = jnp.reshape(emb, (inputs.shape[0], -1, emb.shape[-1]))
      emb = fastmath.dynamic_slice_in_dim(emb, state, inputs.shape[1], axis=1)
      self.state = state + inputs.shape[1]
      return inputs + emb
    elif self._dropout == 0:
      # TODO(kitaev): concat-then-reshape (as is the case with dropout enabled)
      # leads to memory blow-up on TPU.
      # emb = jnp.concatenate(embs, -1)
      # return inputs + jnp.reshape(emb, inputs.shape), state
      return inputs + jnp.concatenate(
          [jnp.reshape(emb, inputs.shape[:-1] + (emb.shape[-1],))
           for emb in embs
          ], -1)
    else:
      emb = jnp.concatenate(embs, -1)
      noise_shape = list(emb.shape)
      for dim in self._dropout_broadcast_dims:
        noise_shape[dim] = 1
      keep_prob = 1.0 - self._dropout
      keep = fastmath.random.bernoulli(rng, keep_prob, tuple(noise_shape))
      multiplier = keep.astype(inputs.dtype) / keep_prob
      return inputs + jnp.reshape(emb * multiplier, inputs.shape)

  def init_weights_and_state(self, input_signature):
    d_feature = input_signature.shape[-1]
    if sum(self._d_embs) != d_feature:
      raise ValueError(
          f'sum(self._d_embs) != d_feature: '
          f'sum({self._d_embs}) vs d_feature: {d_feature}')

    rngs = fastmath.random.split(self.rng, len(self._d_embs))
    weights = []
    for ax, (ax_rng, d_emb) in enumerate(zip(rngs, self._d_embs)):
      ax_shape = [1] * len(self._shape)
      ax_shape[ax] = self._shape[ax]
      ax_shape = (1,) + tuple(ax_shape) + (d_emb,)
      ax_emb = self._kernel_initializer(ax_shape, ax_rng)
      weights.append(ax_emb)

    # State is EMPTY_STATE by default, stays so except for predict mode.
    if self._mode == 'predict':
      self.state = np.array(0, dtype=np.int32)
    self.weights = tuple(weights)


class SinCosPositionalEncoding(layer_base.Layer):
  """Implements the sin-cos positional encoding."""

  def __init__(self, add_offset=2048, dropout=0.0, dropout_broadcast_dims=(-2,),
               start_from_zero_one_in=2, mode='train'):
    """Creates a SinCosPositionalEncoding instance.

    Args:
      add_offset: Maximumnumber to add to positions during training.
      dropout: Probability of *not* adding positional encoding to a sequence
          position.
      dropout_broadcast_dims: Axes along which dropout mask values are
          broadcast rather than individually set at random.
      start_from_zero_one_in: how often to start from 0 during training,
          every one in that many times (e.g., if 4, then it's 25% of the time).
      mode: One of `'train'`, `'eval'`, or `'predict'`.
    """
    super().__init__()
    self._add_offset = add_offset
    if dropout >= 1.0:
      raise ValueError('Dropout rates must be lower than 1.')
    if mode == 'train':
      self._dropout = dropout
    else:
      self._dropout = 0.0
    self._dropout_broadcast_dims = dropout_broadcast_dims
    self._start_from_zero_one_in = start_from_zero_one_in
    self._mode = mode

  def _sincos(self, start, length, d_feature):
    """Create the sin-cos tensor of shape [1, length, d_feature]."""
    position = jnp.arange(0, length)[:, None] + start
    div_term = jnp.exp(
        jnp.arange(0, d_feature, 2) * -(jnp.log(10000.0) / d_feature))
    sin = jnp.sin(position * div_term)
    cos = jnp.cos(position * div_term)
    pe = jnp.concatenate([sin, cos], axis=1)
    return pe[None, :, :]  # [1, length, d_feature]

  def forward(self, inputs):
    """Returns the input activations, with added positional information."""
    if self._mode != 'predict':
      x = inputs
      length = jnp.shape(x)[1]
      if self._mode != 'train':
        start = 0
      else:
        rng1, rng2 = fastmath.random.split(self.rng, 2)
        start = fastmath.random.randint(rng1, (), 0, self._add_offset)
        start_from_nonzero = fastmath.random.randint(
            rng2, (), 0, self._start_from_zero_one_in)
        start_from_nonzero = jnp.minimum(1, start_from_nonzero)
        start *= start_from_nonzero
      px = self._sincos(start, length, inputs.shape[2])
      if self._dropout == 0:
        return x + px
      else:
        noise_shape = list(px.shape)
        for dim in self._dropout_broadcast_dims:
          noise_shape[dim] = 1
        keep_prob = 1.0 - self._dropout
        keep = fastmath.random.bernoulli(self.rng, keep_prob,
                                         tuple(noise_shape))
        multiplier = keep.astype(x.dtype) / keep_prob
        return x + px * multiplier
    else:
      if self._dropout != 0:
        raise ValueError(f'In predict mode, but dropout rate '
                         f'({self._dropout}) is not zero.')

      # State in this class is only used for fast inference. In that case,
      # the model is called with consecutive elements position-by-position.
      # This positional encoding layer needs to store the index of the current
      # position then and increment it on each call -- that's how state is used
      # and updated below.
      pe = self._sincos(self.state, inputs.shape[1], inputs.shape[2])
      self.state += inputs.shape[1]
      return inputs + pe

  def init_weights_and_state(self, input_signature):
    """Randomly initializes the positional encoding vectors.

    Args:
      input_signature: `ShapeDtype` instance characterizing the input this
          layer should compute on.
    """
    if self._mode == 'predict':
      self.state = jnp.zeros((), dtype=jnp.int32)


class FixedBasePositionalEncoding(layer_base.Layer):
  """Implements fixed-base positional encoding."""

  def __init__(self, bases=[11, 13, 14, 15], n_digits=8,  #  pylint: disable=dangerous-default-value
               start_from_zero_one_in=2, base_dropout_one_in=100,
               mode='train', initializer=init.RandomUniformInitializer(1e-4)):
    super().__init__()
    self._bases = bases
    self._n_digits = n_digits
    self._mode = mode
    self._initializer = initializer
    self._start_from_zero_one_in = start_from_zero_one_in
    self._base_dropout_one_in = base_dropout_one_in

  def forward(self, x):
    rng = self.rng
    base_weights, start_vec = self.weights
    batch_size, length = x.shape[0], x.shape[1]
    max_pos = min(self._bases)**self._n_digits
    rng1, rng2, rng3 = fastmath.random.split(rng, 3)
    assert length < max_pos, 'length (%d) >= max_pos (%d)' % (length, max_pos)
    positions = jnp.arange(0, length)[None, :]
    # In training we'll randomize starts for better generalization.
    # We use the trainable start_vec to compensate and give model a way
    # to learn what is the starting position in a sequence.
    if self._mode == 'train':
      # In 1% of training cases still start from 0 to be exactly as in eval.
      start_from_nonzero = fastmath.random.randint(
          rng1, (batch_size,), 0, self._start_from_zero_one_in)
      start_from_nonzero = jnp.minimum(1, start_from_nonzero)
      random_start = fastmath.random.randint(
          rng2, (batch_size,), 0, max_pos-length)
      random_start *= start_from_nonzero
      positions += random_start[:, None]
    if self._mode == 'predict':
      positions += self.state
    res = []
    for bn, base in enumerate(self._bases):
      pos_embeddings = []
      cur_positions = positions
      for i in range(self._n_digits):
        cur_indices = jnp.mod(cur_positions, base)
        cur_positions = cur_positions // base
        s = base_weights[bn][i]
        pos_embeddings.append(cur_indices.astype(jnp.float32)[:, :, None] * s)
      embeddings = jnp.concatenate(pos_embeddings, axis=-1)
      if self._mode == 'train':
        base_dropout = fastmath.random.randint(
            rng3, (batch_size,), 0, self._base_dropout_one_in)
        base_dropout = jnp.minimum(1, base_dropout).astype(jnp.float32)
        embeddings *= base_dropout[:, None, None]
      res.append(embeddings)
    res = sum(res)  # Sum embeddings from all bases.
    # Add start_vec to the first position only to mark it as starting.
    res0 = res[:, 0, :][:, None, :]
    start_pos = res0 + start_vec
    if self._mode == 'predict':
      start_pos = jnp.where(jnp.equal(self.state, 0), start_pos, res0)
      self.state += length  # Add input length to state.
    res = jnp.concatenate([start_pos, res[:, 1:, :]], axis=1)
    return x + res

  def init_weights_and_state(self, input_signature):
    d_feature = input_signature.shape[-1]
    if d_feature % self._n_digits != 0:
      raise ValueError(
          f'd_feature({d_feature}) % self._n_digits({self._n_digits}) != 0')
    d_weight = d_feature // self._n_digits
    rng1, rng2 = fastmath.random.split(self.rng, 2)
    base_weights = [[self._initializer((1, d_weight), rng)
                     for rng in fastmath.random.split(rng1, self._n_digits)]
                    for _ in self._bases]
    # Special vector to mark the starting position.
    start_vec = self._initializer((1, 1, d_feature), rng2)
    self.weights = (base_weights, start_vec)
    if self._mode == 'predict':
      self.state = jnp.zeros((), dtype=jnp.int32)


def threefry_2x32_prf(key, x: jnp.ndarray) -> jnp.ndarray:
  """Apply the threefry PRF to an array of inputs.

  This function is vectorized over x.
  For threefry_2x32: K = X = uint32[2]

  Args:
    key: uint32[2] the key of the PRF
    x: uint32[..., 2] the inputs

  Returns:
    y: uint32[..., 2] the outputs
  """
  if not (key.shape == (2,) and key.dtype == jnp.uint32):
    raise TypeError('key must be uint32[2]', key)
  if not (x.shape[-1:] == (2,) and x.dtype == jnp.uint32):
    raise TypeError('x must be uint32[..., 2]', x)
  # Threefry-2x32 expects this weird format:
  x_3f = jnp.moveaxis(x, source=-1, destination=0).flatten()
  y_3f = jax.random.threefry_2x32(key, x_3f)
  y = jnp.moveaxis(
      jnp.reshape(y_3f, (2,) + x.shape[:-1]), source=0, destination=-1)
  return y


def threefry_2x32_prange(key, lo: int = 0, hi: int = 2):
  """Splits a key into a stream of random keys.

  This uses the little-endian counter mode.

  Args:
    key: uint32[2] the key to split
    lo: the range to start extracting from
    hi: the range to stop extracting from

  Returns:
    keys: uint32[hi - lo, 2] the split keys
  """
  if not (key.shape == (2,) and key.dtype == jnp.uint32):
    raise ValueError('key must be uint32[2]')
  if not hi < 2**32:
    # You shouldn't really be using more than half the key size anyways.
    raise NotImplementedError('only 32-bit sizes are supported')
  # Create a 64-bit counter:
  i_lo = jnp.arange(lo, hi, dtype=jnp.uint32)
  i_hi = jnp.zeros_like(i_lo)
  i = jnp.stack([i_lo, i_hi], axis=-1)
  return threefry_2x32_prf(key, i)


class InfinitePositionalEncoding(layer_base.Layer):
  """Infinite positional encoding."""

  def __init__(
      self, drift=.03, affine=True, transform='any',
      time_bin_length=None,
      mode='train'):
    """Initializes the encoding.

    The encoding tries to roughly evenly traverse the latent space.
    The recurrence time is dependent on how many bits per dimension you use.

    There are two parameters to control randomization:
    - randomizing the origin every 1/drift steps by letting it drift
    - randomizing the origin per call

    Args:
      drift: variance in position difference per unit of difference
      affine: whether to randomize the origin every call
      transform: learnable transform after encoding (any/diag/none)
      time_bin_length: Add features AxialPositionalEncoding learns if
        TimeBinCausalAttention is the first layer.
        bin_length should match TBCA.bin_length
        If you set transform='diag', this flag increases your model capacity to
        close to transform='any', though it will still train slower.
      mode: if 'predict', allow evaluating one token at a time
    """
    super().__init__()
    if transform not in ('any', 'diag', 'none'):
      raise ValueError(transform)
    self._noise_rng = jax.random.split(jax.random.PRNGKey(234234535))[0]
    assert self._noise_rng is not None
    self._noise = None
    self._drift = drift
    self._affine = affine
    self._transform = transform
    self._time_bin_length = time_bin_length
    self._mode = mode

  def _get_noise(self, lo: int, hi: int, depth: int):
    """Return pseudorandom noise with shape float[length, depth].

    Args:
      lo: where to start sampling
      hi: where to stop sampling
      depth: noise depth

    Returns:
      noise[lo:hi, :]: the noise, where noise.diff(axis=0) is i.i.d. U(-1,1)
    """
    if self._noise is None or self._noise.shape[0] < hi:
      # Resize the noise:
      new_length = 1
      while new_length < hi:
        new_length *= 2
      noise = threefry_2x32_prange(self._noise_rng, 0, new_length * depth)
      noise = noise.reshape((new_length, depth, 2))[:, :, 0]
      # Normalize to [-sqrt(3), sqrt(3)]:
      noise = noise.astype(jnp.float32) / np.float32(2**31 - 1)
      noise = noise * 3**.5
      # TODO(tying): use multiscale noise for memory-efficient sampling
      noise = noise.cumsum(axis=0)
      self._noise = noise
    assert self._noise.shape[0] >= hi
    assert self._noise.shape[1] == depth
    return self._noise[lo:hi, :]

  def _get_embeddings(self, lo: int, hi: int, depth, rng=None):
    """Get embeddings float[length, depth].

    Args:
      lo: where to start sampling
      hi: where to stop sampling
      depth: embedding depth
      rng: rng for random phase

    Returns:
      embeddings: float[length, depth]
    """
    noise = self._get_noise(lo, hi, (depth + 1) // 2)
    # Make the stddev around 1 after 1/drift.
    noise = noise * self._drift**.5

    t, c = np.mgrid[lo:hi, :depth]
    # Make even channels cos, odd channels sin:
    c_div_2, c_mod_2 = divmod(c, 2)
    # Off-by-one correction for odd depth:
    drift = self._drift
    if depth > 2:
      drift = drift**(((depth+1)//2)/(depth//2))
    # Spend roughly half the frequencies on noise:
    freq = jnp.geomspace(.5, .5 * drift**2, num=(depth + 1) // 2)[c_div_2]
    cycles = c_mod_2 / 4 + freq * t + noise[:, c_div_2[0, :]] / 4
    assert cycles.shape == (hi - lo, depth), cycles.shape

    # Get random phases:
    if self._affine:
      assert rng is not None
      cycles = cycles + trax.fastmath.random.uniform(
          rng, (1, depth,), minval=0, maxval=1)

    # Convert from cycles to radians:
    embeddings = jnp.cos(jnp.pi * 2 * cycles)

    # Set the last channels to the time bin features:
    if self._time_bin_length is not None:
      inter_bin_idx, intra_bin_idx = divmod(t[:, -1:], self._time_bin_length)
      bin_parity = inter_bin_idx % 2
      bin_fraction = intra_bin_idx / self._time_bin_length
      embeddings = jnp.concatenate(
          [
              embeddings[:, :-3],
              1 / (1 + inter_bin_idx),
              bin_fraction,
              bin_parity.astype(jnp.float32),
          ], -1)

    assert embeddings.shape == (hi - lo, depth), embeddings.shape
    return embeddings

  def forward(self, inputs):
    rng, state = self.rng, self.state
    d_feature = inputs.shape[-1]
    input_len = inputs.shape[-2]

    if self._mode == 'predict':
      # Assume all the positions are pretty close to each other.
      index, predict_rng = state
      lo = index.min()
      hi = index.max() + 1
      emb = self._get_embeddings(lo=lo, hi=hi, depth=d_feature, rng=predict_rng)
      emb = emb[index - lo, jnp.newaxis, :]
      index = index + 1
      state = index, predict_rng
    else:
      emb = self._get_embeddings(lo=0, hi=input_len, depth=d_feature, rng=rng)
      emb = emb[jnp.newaxis, :input_len, :]
    # TODO(tying): check that XLA swaps matmul(slice(x)) -> slice(matmul(x)),
    # or inline this code into get_embeddings/get_noise
    if self._transform == 'diag':
      emb = emb * jax.nn.softplus(self.weights)
    elif self._transform == 'any':
      emb = emb @ self.weights
    self.state = state
    return inputs + emb

  def init_weights_and_state(self, input_signature):
    d_feature = input_signature.shape[-1]
    if self._transform == 'diag':
      # Initialize it to a small value because JAX has a bug in softplus.
      scale_isoftplus = jnp.zeros((d_feature,), dtype=jnp.float32) + 1e-4
      weights = scale_isoftplus
    elif self._transform == 'any':
      ortho = trax.layers.initializers.OrthogonalInitializer()
      weights = ortho((d_feature, d_feature), self.rng)
    else:
      weights = layer_base.EMPTY_WEIGHTS
    if self._mode == 'predict':
      batch_size = input_signature.shape[0]
      self.state = jnp.zeros((batch_size,), dtype=jnp.int32), self.rng
    self.weights = weights


class TimeBinPositionalEncoding(layer_base.Layer):
  """Just the engineered features from InfinitePositionalEncoding."""
  num_features = 3

  def __init__(self, time_bin_length, mode='train'):
    """Initializes the encoding.

    Args:
      time_bin_length: TimeBinCausalAttention.bin_length of the first layer.
      mode: if 'predict', allow evaluating one token at a time
    """
    super().__init__()
    self._time_bin_length = time_bin_length
    self._mode = mode

  def _get_embeddings(self, t):
    """Get embeddings float[..., num_features].

    Args:
      t: int[...] position (i.e. jnp.arange(..., jnp.int32))

    Returns:
      embeddings: float[..., num_features]
    """
    inter_bin_idx, intra_bin_idx = divmod(t, self._time_bin_length)
    bin_parity = inter_bin_idx % 2
    bin_fraction = intra_bin_idx / self._time_bin_length
    embeddings = jnp.stack([
        1 / (1 + inter_bin_idx),
        bin_fraction,
        bin_parity.astype(jnp.float32),
    ], -1)

    assert embeddings.shape == t.shape + (self.num_features,), embeddings.shape
    return embeddings

  def forward(self, inputs):
    state = self.state
    depth = inputs.shape[-1]

    if self._mode == 'predict':
      emb = self._get_embeddings(t=state)
      emb = emb[:, jnp.newaxis, :]
      state = state + 1
    else:
      input_len = inputs.shape[-2]
      emb = self._get_embeddings(t=jnp.arange(input_len, dtype=jnp.int32))
      # Leave batch axis as 1 for broadcasting:
      emb = emb[jnp.newaxis, :, :]
      emb = jnp.broadcast_to(emb, inputs.shape[:-1] + (3,))

    # Replace the last num_features channels of input.
    inputs = jnp.concatenate([inputs[..., :-self.num_features], emb], -1)
    if inputs.shape[-1] > depth:
      logging.warning(
          'dropping feature(s): %d down to %d', inputs.shape[-1], depth)
      inputs = inputs[..., -depth:]

    assert inputs.shape[-1] == depth, inputs.shape
    self.state = state
    return inputs

  def init_weights_and_state(self, input_signature):
    if self._mode == 'predict':
      batch_size = input_signature.shape[0]
      self.state = jnp.zeros((batch_size,), dtype=jnp.int32)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for decoding."""

import functools
import os

import gin
from jax import test_util  # pylint: disable=unused-import
from jax.config import config
import numpy as np
from tensorflow.compat.v2 import test

from trax import fastmath
from trax import layers as tl
from trax import models
from trax import shapes
from trax.supervised import decoding


pkg_dir, _ = os.path.split(__file__)
_TESTDATA = os.path.join(pkg_dir, 'testdata')
_CONFIG_DIR = os.path.join(pkg_dir, 'configs/')


class DecodingTest(test.TestCase):

  def test_autoregressive_sample_transformerlm(self):
    model = models.TransformerLM(10, d_model=32, d_ff=64, n_layers=1,
                                 n_heads=2, mode='predict')
    model.init(shapes.ShapeDtype((1, 1), dtype=np.int32))
    s1 = decoding.autoregressive_sample(
        model, batch_size=1, eos_id=-1, max_length=10)
    self.assertEqual(s1.shape[0], 1)
    self.assertEqual(s1.shape[1], 10)
    batch_per_device = 2 // fastmath.local_device_count()
    model.init(shapes.ShapeDtype((batch_per_device, 1), dtype=np.int32))
    s2 = decoding.autoregressive_sample(
        model, batch_size=2, max_length=10)
    self.assertEqual(s2.shape[0], 2)
    self.assertLess(s2.shape[1], 11)
    model.init(shapes.ShapeDtype((1, 1), dtype=np.int32))
    prefix = np.array([[1, 2, 3]])
    s3 = decoding.autoregressive_sample(model, prefix, eos_id=-1, max_length=10,
                                        batch_size=1)
    self.assertEqual(s3.shape[0], 1)
    self.assertEqual(s3.shape[1], 10)

  def test_autoregressive_sample_transformerlm_tfnp(self):
    with fastmath.use_backend(fastmath.Backend.TFNP):
      model = models.TransformerLM(10, d_model=32, d_ff=64, n_layers=1,
                                   n_heads=2, mode='predict')
      model.init(shapes.ShapeDtype((1, 1), dtype=np.int32))
      s1 = decoding.autoregressive_sample(
          model, batch_size=1, eos_id=-1, max_length=10)
      self.assertEqual(s1.shape[0], 1)
      self.assertEqual(s1.shape[1], 10)
      batch_per_device = 2 // fastmath.local_device_count()
      model.init(shapes.ShapeDtype((batch_per_device, 1), dtype=np.int32))
      s2 = decoding.autoregressive_sample(
          model, batch_size=2, max_length=10)
      self.assertEqual(s2.shape[0], 2)
      self.assertLess(s2.shape[1], 11)
      model.init(shapes.ShapeDtype((1, 1), dtype=np.int32))
      prefix = np.array([[1, 2, 3]])
      s3 = decoding.autoregressive_sample(model, prefix, eos_id=-1,
                                          max_length=10, batch_size=1)
      self.assertEqual(s3.shape[0], 1)
      self.assertEqual(s3.shape[1], 10)

  def _lsh_self_attention_fn(self):
    return functools.partial(
        tl.LSHSelfAttention,
        attention_dropout=0.0,
        chunk_len=64,
        n_buckets=[32, 32],
        n_chunks_after=0,
        n_chunks_before=1,
        n_hashes=1,
        n_parallel_heads=1,
        predict_drop_len=128,
        predict_mem_len=1024,
    )

  def _pure_lsh_self_attention_fn(self, n_chunks_after=0):
    return functools.partial(
        tl.PureLSHSelfAttentionWrapper,
        attention_dropout=0.0,
        chunk_len=16,
        n_buckets=[32, 32],
        n_chunks_after=n_chunks_after,
        n_chunks_before=1,
        n_hashes=2,
        n_parallel_heads=1,
        max_length_for_buckets=1024,
        predict_drop_len=128,
        predict_mem_len=1024,
        num_weights=2,
        bias=False,
        pure_lsh_implementation=tl.PureLSHSelfAttention,
    )

  def _timebin_self_attention_fn(self, use_reference_code=False, chunk_len=64):
    return functools.partial(
        tl.SelfAttention,
        attention_dropout=0.05,
        chunk_len=chunk_len,
        n_chunks_before=1,
        n_parallel_heads=1,
        use_reference_code=use_reference_code,
        predict_drop_len=128,
        predict_mem_len=1024,
    )

  def test_autoregressive_sample_reformerlm(self):
    lsh_self_attention = self._lsh_self_attention_fn()
    timebin_self_attention = self._timebin_self_attention_fn()

    model = models.ReformerLM(vocab_size=256,
                              d_model=256,
                              d_ff=512,
                              d_attention_key=128,
                              d_attention_value=128,
                              n_layers=2,
                              n_heads=2,
                              dropout=0.05,
                              max_len=65536,
                              attention_type=[timebin_self_attention,
                                              lsh_self_attention],
                              pos_axial_shape=(256, 256),
                              pos_d_axial_embs=(128, 128),
                              ff_activation=tl.Relu,
                              ff_use_sru=0,
                              mode='predict',
                              )
    model.init(shapes.ShapeDtype((1, 1), dtype=np.int32))
    s1 = decoding.autoregressive_sample(
        model, batch_size=1, eos_id=-1, max_length=10)
    self.assertEqual(s1.shape[0], 1)
    self.assertEqual(s1.shape[1], 10)

  def test_autoregressive_sample_transformer(self):
    model = models.Transformer(10, d_model=32, d_ff=64, n_encoder_layers=1,
                               n_decoder_layers=1, n_heads=2, mode='predict')
    inputs = np.ones((1, 3), dtype=np.int32)
    model.init((shapes.signature(inputs),
                shapes.ShapeDtype((1, 1), dtype=np.int32)))
    s = decoding.autoregressive_sample(model, inputs=inputs,
                                       eos_id=-1, max_length=10)
    self.assertEqual(s.shape[0], 1)
    self.assertEqual(s.shape[1], 10)

  def test_autoregressive_sample_transformerlm_quality(self):
    pred_model = models.TransformerLM(
        d_model=64, d_ff=128, dropout=0.05, max_len=256, n_heads=2,
        n_layers=2, vocab_size=13, mode='predict')
    shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
    model_path = os.path.join(_TESTDATA, 'transformerlm_copy.pkl.gz')
    pred_model.init_from_file(model_path, weights_only=True,
                              input_signature=(shape11, shape11))
    inputs = np.array([[0, 3, 7, 5, 3, 2, 4, 0]], dtype=np.int32)
    s = decoding.autoregressive_sample(pred_model, inputs,
                                       max_length=6, temperature=0.0)
    self.assertEqual(str(s[0]), '[3 7 5 3 2 4]')

  def test_autoregressive_sample_transformerlm_quality_eval(self):
    eval_model = models.TransformerLM(
        d_model=64, d_ff=128, dropout=0.05, max_len=256, n_heads=2,
        n_layers=2, vocab_size=13, mode='eval')
    model_path = os.path.join(_TESTDATA, 'transformerlm_copy.pkl.gz')
    eval_model.init_from_file(model_path)
    inputs = np.array([[0, 3, 7, 5, 3, 2, 4, 0]], dtype=np.int32)
    s = decoding.autoregressive_sample(eval_model, inputs, eval_mode=True,
                                       max_length=6, temperature=0.0)
    self.assertEqual(str(s[0]), '[3 7 5 3 2 4]')

  def test_autoregressive_sample_transformerlm_quality_beam(self):
    pred_model = models.TransformerLM(
        d_model=64, d_ff=128, dropout=0.05, max_len=256, n_heads=2,
        n_layers=2, vocab_size=13, mode='predict')
    shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
    model_path = os.path.join(_TESTDATA, 'transformerlm_copy.pkl.gz')
    pred_model.init_from_file(model_path, weights_only=True,
                              input_signature=(shape11, shape11))
    inputs = np.array([[0, 3, 7, 5, 3, 2, 4, 0]], dtype=np.int32)
    s = decoding.beam_search(pred_model, inputs, n_beams=3, max_length=6)
    self.assertEqual(len(s), 3)  # 3 beams
    self.assertEqual(str(s[0][0][0]), '[3 7 5 3 2 4]')
    self.assertEqual(str(s[1][0][0]), '[3 7 5 3 2 2]')  # different from above
    self.assertEqual(str(s[2][0][0]), '[3 7 5 3 3 2]')  # different from above

  def test_autoregressive_sample_transformer_quality(self):
    pred_model = models.Transformer(
        d_model=64, d_ff=128, dropout=0.05, max_len=256, n_heads=2,
        n_encoder_layers=2, n_decoder_layers=2, input_vocab_size=13,
        mode='predict')
    shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
    model_path = os.path.join(_TESTDATA, 'transformer_copy.pkl.gz')
    pred_model.init_from_file(model_path, weights_only=True,
                              input_signature=(shape11, shape11))
    inputs = np.array([[3, 7, 5, 3, 2, 4, 1, 8]], dtype=np.int32)
    s = decoding.autoregressive_sample(pred_model, inputs=inputs,
                                       eos_id=1, max_length=10, temperature=0.0)
    self.assertEqual(str(s[0]), '[3 7 5 3 2 4 1]')

  def test_autoregressive_sample_terraformer_lsh(self):
    max_len = 128

    pred_model = models.ConfigurableTerraformer(
        mode='predict',
        d_model=256,
        d_ff=512,
        dropout=0.05,
        max_len=max_len,
        n_heads=4,
        n_encoder_layers=1,
        n_decoder_layers=1,
        ff_use_sru=1,
        d_attention_key=64,
        d_attention_value=64,
        encoder_attention_type=self._lsh_self_attention_fn(),
        encoder_decoder_attention_type=self._lsh_self_attention_fn(),
        input_vocab_size=256,
        pos_axial_shape=None,
    )

    shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
    shape1l = shapes.ShapeDtype((1, max_len), dtype=np.int32)
    pred_model.init(input_signature=(shape1l, shape11))

    # 0w0w
    inputs = np.array(
        [[0, 3, 7, 5, 3, 2, 4, 1, 8, 0, 3, 7, 5, 3, 2, 4, 1, 8]],
        dtype=np.int32)
    inputs = np.pad(inputs, [(0, 0), (0, max_len - inputs.shape[1])],
                    mode='constant', constant_values=0)
    s = decoding.autoregressive_sample(
        pred_model, inputs=inputs, eos_id=-1, max_length=10, temperature=0.0)

    self.assertEqual(s.shape[0], 1)
    self.assertEqual(s.shape[1], 10)

  def test_autoregressive_sample_terraformer_lsh_attn_quality(self):
    gin.add_config_file_search_path(_CONFIG_DIR)
    max_len = 32  # 32 is the max length we trained the checkpoint for.
    test_lengths = [8, 16, 32]
    vocab_size = 13
    # The checkpoint is correct on ~90% sequences, set random seed to deflake.
    np.random.seed(0)
    for test_len in test_lengths:
      gin.clear_config()
      gin.parse_config_file('terraformer_copy.gin')
      gin.bind_parameter('LSHSelfAttention.predict_mem_len', 2 * max_len)
      gin.bind_parameter('LSHSelfAttention.predict_drop_len', 2 * max_len)

      pred_model = models.ConfigurableTerraformer(mode='predict')

      shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
      shape1l = shapes.ShapeDtype((1, max_len), dtype=np.int32)

      model_path = os.path.join(_TESTDATA, 'terraformer_copy_lsh_attn.pkl.gz')
      pred_model.init_from_file(model_path, weights_only=True,
                                input_signature=(shape1l, shape11))
      initial_state = pred_model.state

      for _ in range(2):  # Set low to make the test run reasonably fast.
        # Pick a length in [1, test_len] at random.
        inp_len = np.random.randint(low=1, high=test_len + 1)
        inputs = np.random.randint(low=1, high=vocab_size-1, size=(1, max_len))
        # TODO(jaszczur): properly fix padding in terraformer predict mode,
        # and add a test here.
        s = decoding.autoregressive_sample(
            pred_model, inputs=inputs, eos_id=-1, max_length=inp_len,
            temperature=0.0)
        np.testing.assert_equal(s[0], inputs[0, :inp_len])
        pred_model.state = initial_state
    gin.clear_config()  # Make sure to not affect other tests.

  def test_autoregressive_sample_reformerlm_lsh(self):
    max_len = 32

    pred_model = models.ReformerLM(
        mode='predict',
        d_model=256,
        d_ff=512,
        dropout=0.05,
        max_len=2 * max_len,
        n_heads=4,
        n_layers=3,
        ff_use_sru=0,
        d_attention_key=64,
        d_attention_value=64,
        attention_type=functools.partial(tl.LSHSelfAttention,
                                         chunk_len=16,
                                         n_hashes=2,
                                         n_buckets=[32, 32],
                                         predict_drop_len=max_len,
                                         predict_mem_len=max_len,
                                         max_length_for_buckets=1024),
        vocab_size=13,
        pos_type='fixed-base',
        pos_d_axial_embs=None,
    )

    shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
    pred_model.init(shape11)

    # 0w0
    inputs = np.array([[0, 3, 7, 5, 3, 2, 0]], dtype=np.int32)
    inputs = np.pad(inputs, [(0, 0), (0, max_len - inputs.shape[1])],
                    mode='constant', constant_values=0)
    s = decoding.autoregressive_sample(
        pred_model, inputs=inputs, eos_id=-1, max_length=10, temperature=0.0)

    self.assertEqual(s.shape[0], 1)
    self.assertEqual(s.shape[1], 10)

  def test_autoregressive_sample_reformerlm_lsh_quality(self):
    max_len = 32

    pred_model = models.ReformerLM(
        mode='predict',
        d_model=256,
        d_ff=512,
        dropout=0.05,
        max_len=2 * max_len,
        n_heads=4,
        n_layers=3,
        ff_use_sru=0,
        d_attention_key=64,
        d_attention_value=64,
        attention_type=functools.partial(tl.LSHSelfAttention,
                                         chunk_len=16,
                                         n_hashes=2,
                                         n_buckets=[32, 32],
                                         predict_drop_len=max_len,
                                         predict_mem_len=max_len,
                                         max_length_for_buckets=1024),
        vocab_size=13,
        pos_type='fixed-base',
        pos_d_axial_embs=None,
    )

    shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)

    model_path = os.path.join(
        _TESTDATA, 'reformerlm_copy_lsh_attn.pkl.gz')
    pred_model.init_from_file(model_path, weights_only=True,
                              input_signature=shape11)

    # 0w0
    inputs = np.array([[0, 3, 7, 5, 3, 2, 0]], dtype=np.int32)
    inp_len = inputs.shape[1]
    s = decoding.autoregressive_sample(
        pred_model, inputs=inputs, eos_id=-1, max_length=inp_len-2,
        temperature=0.0)

    np.testing.assert_equal(s[0], inputs[0, 1:inp_len-1])
    # pylint: enable=unreachable

  def test_autoregressive_sample_terraformer_pure_lsh(self):
    max_len = 128

    pred_model = models.ConfigurableTerraformer(
        mode='predict',
        d_model=256,
        d_ff=512,
        dropout=0.05,
        max_len=max_len,
        n_heads=4,
        n_encoder_layers=1,
        n_decoder_layers=1,
        ff_use_sru=1,
        d_attention_key=64,
        d_attention_value=64,
        encoder_attention_type=self._pure_lsh_self_attention_fn(
            n_chunks_after=1),
        encoder_decoder_attention_type=self._pure_lsh_self_attention_fn(),
        input_vocab_size=256,
        pos_axial_shape=None,
    )

    shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
    shape1l = shapes.ShapeDtype((1, max_len), dtype=np.int32)
    pred_model.init(input_signature=(shape1l, shape11))

    # 0w0w
    inputs = np.array(
        [[0, 3, 7, 5, 3, 2, 4, 1, 8, 0, 3, 7, 5, 3, 2, 4, 1, 8]],
        dtype=np.int32)
    inputs = np.pad(inputs, [(0, 0), (0, max_len - inputs.shape[1])],
                    mode='constant', constant_values=0)
    s = decoding.autoregressive_sample(
        pred_model, inputs=inputs, eos_id=-1, max_length=10, temperature=0.0)

    self.assertEqual(s.shape[0], 1)
    self.assertEqual(s.shape[1], 10)

  def test_autoregressive_sample_terraformer_pure_lsh_attn_quality(self):
    gin.add_config_file_search_path(_CONFIG_DIR)
    max_len = 32  # 32 is the max length we trained the checkpoint for.
    test_lengths = [8, 16, 32]
    vocab_size = 13
    # The checkpoint is correct on ~90% sequences, set random seed to deflake.
    np.random.seed(0)
    for test_len in test_lengths:
      gin.clear_config()
      gin.parse_config_file('terraformer_purelsh_copy.gin')
      gin.bind_parameter('PureLSHSelfAttention.predict_mem_len', 2 * max_len)
      gin.bind_parameter('PureLSHSelfAttention.predict_drop_len', 2 * max_len)
      gin.bind_parameter('PureLSHSelfAttentionWrapper.bias', False)
      gin.bind_parameter('PureLSHSelfAttentionWrapper.num_weights', 2)

      pred_model = models.ConfigurableTerraformer(mode='predict')

      shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
      shape1l = shapes.ShapeDtype((1, max_len), dtype=np.int32)

      model_path = os.path.join(_TESTDATA, 'terraformer_purelsh_copy.pkl.gz')
      pred_model.init_from_file(model_path, weights_only=True,
                                input_signature=(shape1l, shape11))
      initial_state = pred_model.state

      for _ in range(2):  # Set low to make the test run reasonably fast.
        # Pick a length in [1, test_len] at random.
        inp_len = np.random.randint(low=1, high=test_len + 1)
        inputs = np.random.randint(low=1, high=vocab_size-1, size=(1, max_len))
        # TODO(jaszczur): properly fix padding in terraformer predict mode,
        # and add a test here.
        s = decoding.autoregressive_sample(
            pred_model, inputs=inputs, eos_id=-1, max_length=inp_len,
            temperature=0.0)

        np.testing.assert_equal(s[0], inputs[0, :inp_len])
        pred_model.state = initial_state
    gin.clear_config()  # Make sure to not affect other tests.


if __name__ == '__main__':
  config.config_with_absl()
  test.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for RL training."""

import functools

from absl.testing import absltest

from trax import models
from trax import optimizers as opt
from trax.models import atari_cnn
from trax.rl import actor_critic
from trax.rl import task as rl_task
from trax.supervised import lr_schedules




if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trax fast math: pure numpy backend."""

import numpy as np
from scipy.special import logsumexp
from trax.shapes import signature


def get_prng(seed):
  """JAX-compatible way of getting PRNG seeds."""
  if np.shape(seed):
    raise TypeError('PRNGKey seed must be a scalar.')
  convert = lambda k: np.reshape(np.asarray(k, np.uint32), [1])
  k1 = convert(np.bitwise_and(np.right_shift(seed, 32), 0xFFFFFFFF))
  k2 = convert(np.bitwise_and(seed, 0xFFFFFFFF))
  return np.concatenate([k1, k2], 0)


def random_uniform(rng, shape=(), dtype=np.float64, minval=0.0, maxval=1.0):
  del rng
  return np.random.uniform(minval, maxval, size=shape).astype(dtype)


def random_normal(rng, shape=(), dtype=np.float64):
  del rng
  return np.random.normal(size=shape).astype(dtype)


def random_randint(rng, shape, minval, maxval, dtype=np.int64):
  del rng
  return np.random.randint(minval, maxval, size=shape).astype(dtype)


def random_bernoulli(rng, p=0.5, shape=()):
  del rng
  return np.random.binomial(1, p, size=shape)


def np_abstract_eval(f):
  """Abstract evaluation in numpy by running the real function on 0s."""
  def abstract_f(*args, **kwargs):
    real_args = [nested_map(lambda x: np.zeros(x.shape, x.dtype), a)
                 for a in args]
    real_res = f(*real_args, **kwargs)
    return signature(real_res)
  return abstract_f


NUMPY_BACKEND = {
    'abstract_eval': np_abstract_eval,
    'local_device_count': lambda: 1,
    'global_device_count': lambda: 1,
    'jit': lambda f: f,
    'logsumexp': logsumexp,
    'name': 'numpy',
    'np': np,
    'random_bernoulli': random_bernoulli,
    'random_get_prng': get_prng,
    'random_normal': random_normal,
    'random_randint': random_randint,
    'random_split': lambda prng, num=2: (None,) * num,
    'random_uniform': random_uniform,
    'expit': lambda x: 1. / (1. + np.exp(-x)),
}


def nested_map(f, obj, level=0, ignore_nones=True):
  """Maps `f` recursively inside any dicts/lists/tuples in `obj`.

  Args:
    f: A function taking a single object as input. f's input must NOT be a
        dict, list, or tuple, or any subclass of those.
    obj: Either an input object to f or some nested structure of collections
        of (collections of ...) input objects to f.
    level: Level in the nested structure to stop at, counted from the leaves -
        so level 0 is the leaf, level 1 is such that all of its children are at
        level 0 etc.
    ignore_nones: Whether to ignore Nones in the structure, i.e. return None
        without calling `f`.

  Returns:
    An object with the same nested structure as `obj`, but with each input
    object `x` replaced by `f(x)`.
  """
  if _is_at_level(obj, level):
    if ignore_nones and _is_made_of_nones(obj):
      return None
    else:
      return f(obj)

  if _is_namedtuple_instance(obj):
    return type(obj)(*nested_map(f, list(obj), level=level))
  if isinstance(obj, list):
    return [nested_map(f, y, level=level) for y in obj]
  if isinstance(obj, tuple):
    return tuple([nested_map(f, y, level=level) for y in obj])
  if isinstance(obj, dict):
    return {k: nested_map(f, v, level=level) for (k, v) in obj.items()}

  raise ValueError('Non-exhaustive pattern match for {}.'.format(obj))


def nested_map_multiarg(f, *objs, ignore_nones=True):
  """Maps multi-arg `f` recursively inside any dicts/lists/tuples in `objs`.

  Args:
    f: A function taking len(objs) inputs. f's input must NOT be a
        dict, list, or tuple, or any subclass of those.
    *objs: Either input objects to f or some nested structure of collections
        of (collections of ...) input objects to f.
    ignore_nones: Whether to ignore Nones in the structure, i.e. return None
        without calling `f`.

  Returns:
    An object with the same nested structure as `objs[0]`, but with each input
    object `x` replaced by `f(*xs)`.
  """
  if isinstance(objs[0], list):
    return [nested_map_multiarg(f, *[o[i] for o in objs])
            for i in range(len(objs[0]))]
  if isinstance(objs[0], tuple):
    return tuple([nested_map_multiarg(f, *[o[i] for o in objs])
                  for i in range(len(objs[0]))])
  if isinstance(objs[0], dict):
    return {k: nested_map_multiarg(f, *[o[k] for o in objs])
            for k in objs[0]}
  if ignore_nones and _is_made_of_nones(objs):
    return None
  return f(*objs)


def nested_zip(objs):
  """Zips the leaves of each nested structure in `objs`.

  Args:
    objs: List of nested structures to zip.

  Returns:
    An object with the same nested structure as each element of `objs`, with
    leaves zipped together into tuples.
  """
  assert isinstance(objs, (list, tuple))
  assert objs, 'Cannot zip an empty sequence.'

  if _is_at_level(objs, 1):
    return tuple(objs)

  if _is_namedtuple_instance(objs[0]):
    return type(objs[0])(*nested_zip(list(map(list, objs))))
  if isinstance(objs[0], list):
    return [nested_zip([obj[i] for obj in objs]) for i in range(len(objs[0]))]
  if isinstance(objs[0], tuple):
    return nested_zip(list(map(list, objs)))
  if isinstance(objs[0], dict):
    return {k: nested_zip([obj[k] for obj in objs]) for k in objs[0]}

  raise ValueError('Non-exhaustive pattern match for {}.'.format(objs[0]))


def nested_stack(objs, axis=0, np_module=np):
  """Stacks the numpy arrays inside any dicts/lists/tuples in `objs`.

  Args:
    objs: List of nested structures to stack.
    axis: Axis to stack along.
    np_module: numpy module to use - typically numpy or jax.numpy.

  Returns:
    An object with the same nested structure as each element of `objs`, with
    leaves stacked together into numpy arrays. Nones are propagated, i.e. if
    each element of the stacked sequence is None, the output will be None.
  """
  # nested_map the stacking operation, but stopping at level 1 so at tuples of
  # numpy arrays.
  return nested_map(
      lambda x: np_module.stack(x, axis=axis),
      nested_zip(objs),
      level=1,
  )


def tree_flatten(tree):
  """Flatten a tree into a list."""
  if isinstance(tree, (list, tuple)):
    # In python, sum of lists starting from [] is the concatenation.
    return sum([tree_flatten(t) for t in tree], [])
  if isinstance(tree, dict):
    # Only use the values in case of a dictionary node.
    return sum([tree_flatten(v) for v in tree.values()], [])
  return [tree]


def tree_leaves(tree, ignore_nones=True):
  """Gets the leaves of a tree."""

  # Right now this is just `tree_flatten`, but we keep this separate since
  # JAX's tree_flatten returns the structure of the tree as well.
  flattened = tree_flatten(tree)
  return [flat for flat in flattened if (not ignore_nones) or flat is not None]


def tree_unflatten(flat, tree, copy_from_tree=None):
  """Unflatten a list into a tree given the tree shape as second argument.

  Args:
    flat: a flat list of elements to be assembled into a tree.
    tree: a tree with the structure we want to have in the new tree.
    copy_from_tree: optional list of elements that we just copy from tree.
      This argument is used when the flat version does not contain all elements
      of the expected tree but just a subset, while the rest are filled from
      the tree itself. It allows to omit "unnecessary" elements. For example,
      consider trees (A, (B, X), X) and (X, (A, X), B) where X is some element
      we do not care about. Flattening the first tree and removing X will yield
      a flat list [A, B] and the second tree can then be reconstructed from this
      list and the tree (X, (E, X), E) with copy_from_tree=[X]. One example
      where this is used is the weights-tree of a model, where layers with no
      weights have () in the tree and we use copy_from_tree=[()] to restore
      a model from a file that only has a list of trainable weights.

  Returns:
    A pair (new_tree, rest_of_flat) where the new tree that has the structure
    of tree but with leaves from flat, and the remaining elements of flat if
    more were provided than the number of leaves of tree (useful for recursion).
  """
  if copy_from_tree is not None and tree in copy_from_tree:
    return tree, flat
  if isinstance(tree, (list, tuple)):
    new_tree, rest = [], flat
    for t in tree:
      new_t, rest = tree_unflatten(rest, t, copy_from_tree=copy_from_tree)
      new_tree.append(new_t)
    new_tree = tuple(new_tree) if isinstance(tree, tuple) else new_tree
    return new_tree, rest
  if isinstance(tree, dict):
    new_tree, rest = {}, flat
    for k in tree:
      new_v, rest = tree_unflatten(rest, tree[k], copy_from_tree=copy_from_tree)
      new_tree[k] = new_v
    return new_tree, rest
  return flat[0], flat[1:]


def _is_namedtuple_instance(x):
  """Checks if `x` is an instance of a `namedtuple` type."""
  if not isinstance(x, tuple):
    return False
  return hasattr(x, '_fields')


def _is_at_level(obj, level):
  """Checks if `obj` is an at level `level`."""
  is_leaf = not isinstance(obj, (list, tuple, dict))
  if level == 0 or is_leaf:
    return (level == 0) == is_leaf

  if isinstance(obj, dict):
    elems = obj.values()
  else:
    elems = obj
  return elems and all(_is_at_level(x, level - 1) for x in elems)


def _is_made_of_nones(obj):
  """Checks if `obj` is a nested structure of `None`s."""
  elems = tree_flatten(obj)
  # Returning False for an empty list, because it doesn't have any Nones inside.
  return elems and all(x is None for x in elems)
